"""
Tests for the TemporaryPatchController
"""

# Local
from oper8.patch import STRATEGIC_MERGE_PATCH
from oper8.temporary_patch.temporary_patch_component import TemporaryPatchComponent
from oper8.temporary_patch.temporary_patch_controller import TemporaryPatchController
from oper8.test_helpers.helpers import (
    DummyController,
    MockDeployManager,
    make_patch,
    setup_session,
)

## Tests (non finalier) ########################################################


def test_non_finalizer_patchable_kind_only():
    """Test that when run as a non-finalizer and a patchable kind is given by
    kind only, the component is correctly added to the session
    """
    target = {
        "kind": "Foo",
        "apiVersion": "foo.bar/v2",
        "metadata": {"name": "foo", "namespace": "ns"},
    }
    patch = make_patch(
        patch_type=STRATEGIC_MERGE_PATCH,
        body={"foo": {"key": "value"}},
        target=target,
        namespace="ns",
    )
    dm = MockDeployManager(resources=[target])
    ctrlr = TemporaryPatchController(patchable_kinds=["Foo", "Bar"])
    session = setup_session(full_cr=patch, deploy_manager=dm)
    ctrlr.setup_components(session)
    comps = session.get_components()
    assert len(comps) == 1
    assert isinstance(comps[0], TemporaryPatchComponent)
    assert not session.get_components(disabled=True)


def test_construct_with_controller():
    """Test that constructing with a Controller type as a patchable kind works"""
    ctrlr = TemporaryPatchController(patchable_kinds=[DummyController])
    assert ctrlr._patchable_kinds == [
        "/".join(
            [
                DummyController.group,
                DummyController.version,
                DummyController.kind,
            ]
        )
    ]


def test_non_finalizer_patchable_kind_with_api_version():
    """Test that when run as a non-finalizer and a patchable kind is given by
    kind and api version, the component is correctly added to the session
    """
    target = {
        "kind": "Foo",
        "apiVersion": "foo.bar/v2",
        "metadata": {"name": "foo", "namespace": "ns"},
    }
    patch = make_patch(
        patch_type=STRATEGIC_MERGE_PATCH,
        body={"foo": {"key": "value"}},
        target=target,
        namespace="ns",
    )
    dm = MockDeployManager(resources=[target])
    ctrlr = TemporaryPatchController(patchable_kinds=["foo.bar/v2/Foo"])
    session = setup_session(full_cr=patch, deploy_manager=dm)
    ctrlr.setup_components(session)
    comps = session.get_components()
    assert len(comps) == 1
    assert isinstance(comps[0], TemporaryPatchComponent)
    assert not session.get_components(disabled=True)


def test_non_finalizer_non_patchable_kind_only():
    """Test that when run as a non-finalizer and a patchable kind is given by
    kind only that is not in the list, the component is not added to the session
    """
    target = {
        "kind": "Foo",
        "apiVersion": "foo.bar/v2",
        "metadata": {"name": "foo", "namespace": "ns"},
    }
    patch = make_patch(
        patch_type=STRATEGIC_MERGE_PATCH,
        body={"foo": {"key": "value"}},
        target=target,
        namespace="ns",
    )
    dm = MockDeployManager(resources=[target])
    ctrlr = TemporaryPatchController(patchable_kinds=["Bar"])
    session = setup_session(full_cr=patch, deploy_manager=dm)
    ctrlr.setup_components(session)
    assert not session.get_components()
    assert not session.get_components(disabled=True)


def test_non_finalizer_non_patchable_kind_with_api_version():
    """Test that when run as a non-finalizer and a patchable kind is given by
    kind and api version that is not in the list, the component is not added to
    the session
    """
    target = {
        "kind": "Foo",
        "apiVersion": "foo.bar/v2",
        "metadata": {"name": "foo", "namespace": "ns"},
    }
    patch = make_patch(
        patch_type=STRATEGIC_MERGE_PATCH,
        body={"foo": {"key": "value"}},
        target=target,
        namespace="ns",
    )
    dm = MockDeployManager(resources=[target])
    ctrlr = TemporaryPatchController(patchable_kinds=["foo.bar/v1/Foo"])
    session = setup_session(full_cr=patch, deploy_manager=dm)
    ctrlr.setup_components(session)
    assert not session.get_components()
    assert not session.get_components(disabled=True)


def test_derived_properties():
    """Test that a derived TemporaryPatchController subclass can override the
    desired properties and inherit the non-overridden properties
    """

    class DerivedTempPatchCtrlr(TemporaryPatchController):
        group = "foo.bar.com"

    assert DerivedTempPatchCtrlr.group == "foo.bar.com"
    assert DerivedTempPatchCtrlr.version == "v1"
    assert DerivedTempPatchCtrlr.kind == "TemporaryPatch"
    assert DerivedTempPatchCtrlr.disable_vcs == True


## Tests (finalizer) ###########################################################


def test_finalizer_patchable_kind_only():
    """Test that when run as a finalizer and a patchable kind is given by kind
    only, the component is correctly added to the session
    """
    target = {
        "kind": "Foo",
        "apiVersion": "foo.bar/v2",
        "metadata": {"name": "foo", "namespace": "ns"},
    }
    patch = make_patch(
        patch_type=STRATEGIC_MERGE_PATCH,
        body={"foo": {"key": "value"}},
        target=target,
        namespace="ns",
    )
    dm = MockDeployManager(resources=[target])
    ctrlr = TemporaryPatchController(patchable_kinds=["Foo", "Bar"])
    session = setup_session(full_cr=patch, deploy_manager=dm)
    ctrlr.finalize_components(session)
    comps = session.get_components(disabled=True)
    assert len(comps) == 1
    assert isinstance(comps[0], TemporaryPatchComponent)
    assert not session.get_components(disabled=False)


def test_finalizer_patchable_kind_with_api_version():
    """Test that when run as a finalizer and a patchable kind is given by kind
    and api version, the component is correctly added to the session
    """
    target = {
        "kind": "Foo",
        "apiVersion": "foo.bar/v2",
        "metadata": {"name": "foo", "namespace": "ns"},
    }
    patch = make_patch(
        patch_type=STRATEGIC_MERGE_PATCH,
        body={"foo": {"key": "value"}},
        target=target,
        namespace="ns",
    )
    dm = MockDeployManager(resources=[target])
    ctrlr = TemporaryPatchController(patchable_kinds=["foo.bar/v2/Foo"])
    session = setup_session(full_cr=patch, deploy_manager=dm)
    ctrlr.finalize_components(session)
    comps = session.get_components(disabled=True)
    assert len(comps) == 1
    assert isinstance(comps[0], TemporaryPatchComponent)
    assert not session.get_components(disabled=False)


def test_finalizer_non_patchable_kind_only():
    """Test that when run as a finalizer and a patchable kind is given by kind
    only that is not in the list, the component is not added to the session
    """
    target = {
        "kind": "Foo",
        "apiVersion": "foo.bar/v2",
        "metadata": {"name": "foo", "namespace": "ns"},
    }
    patch = make_patch(
        patch_type=STRATEGIC_MERGE_PATCH,
        body={"foo": {"key": "value"}},
        target=target,
        namespace="ns",
    )
    dm = MockDeployManager(resources=[target])
    ctrlr = TemporaryPatchController(patchable_kinds=["Bar"])
    session = setup_session(full_cr=patch, deploy_manager=dm)
    ctrlr.finalize_components(session)
    assert not session.get_components(disabled=True)
    assert not session.get_components(disabled=False)


def test_finalizer_non_patchable_kind_with_api_version():
    """Test that when run as a finalizer and a patchable kind is given by kind
    and api version that is not in the list, the component is not added to the
    session
    """
    target = {
        "kind": "Foo",
        "apiVersion": "foo.bar/v2",
        "metadata": {"name": "foo", "namespace": "ns"},
    }
    patch = make_patch(
        patch_type=STRATEGIC_MERGE_PATCH,
        body={"foo": {"key": "value"}},
        target=target,
        namespace="ns",
    )
    dm = MockDeployManager(resources=[target])
    ctrlr = TemporaryPatchController(patchable_kinds=["foo.bar/v1/Foo"])
    session = setup_session(full_cr=patch, deploy_manager=dm)
    ctrlr.finalize_components(session)
    assert not session.get_components(disabled=True)
    assert not session.get_components(disabled=False)



"""Annotation Based Leadership Manager"""

# Standard
from copy import copy
from datetime import datetime
from typing import Optional

# First Party
import alog

# Local
from .... import config
from ....constants import LEASE_NAME_ANNOTATION_NAME, LEASE_TIME_ANNOTATION_NAME
from ....deploy_manager import DeployManagerBase
from ....exceptions import assert_config
from ....managed_object import ManagedObject
from ..utils import get_pod_name, parse_time_delta
from .base import LeadershipManagerBase

log = alog.use_channel("LDRELC")


class AnnotationLeadershipManager(LeadershipManagerBase):
    """
    Annotation based leadership manager that uses two annotations
    to track leadership on a per-resource basis. This allows for
    horizontally scalable operations.

    EXPERIMENTAL: This has passed basic validation but has not been rigorously tested
     in the field
    """

    def __init__(self, deploy_manager: DeployManagerBase = None):
        """Initialize Leadership and gather current name

        Args:
            deploy_manager: DeployManagerBase = None
                DeployManager for this Manager
        """

        super().__init__(deploy_manager)
        self.duration_delta = parse_time_delta(
            config.python_watch_manager.lock.duration
        )

        # Gather lock_name, namespace and pod manifest
        self.pod_name = get_pod_name()
        assert_config(self.pod_name, "Unable to detect pod name")

    ## Lock Interface ####################################################
    def acquire(self, force: bool = False) -> bool:
        """
        Return true as leadership is managed at resource level
        """
        return True

    def acquire_resource(self, resource: ManagedObject):
        """Check a resource for leadership annotation and add one if it's expired
        or does not exit"""
        success, current_resource = self.deploy_manager.get_object_current_state(
            resource.kind, resource.name, resource.namespace, resource.api_version
        )
        if not success or not current_resource:
            log.warning(
                "Unable to fetch owner resource %s/%s/%s/%s",
                resource.kind,
                resource.api_version,
                resource.namespace,
                resource.name,
            )
            return False

        if "annotations" not in current_resource.get("metadata"):
            current_resource["metadata"]["annotations"] = {}

        # Check the current annotation
        annotations = current_resource["metadata"]["annotations"]
        current_time = datetime.now()

        # If no leader than take ownership
        if not annotations.get(LEASE_NAME_ANNOTATION_NAME):
            annotations[LEASE_NAME_ANNOTATION_NAME] = self.pod_name
            annotations[LEASE_TIME_ANNOTATION_NAME] = current_time.isoformat()

        # If already the current leader then update lease time
        elif self.pod_name == annotations.get(LEASE_NAME_ANNOTATION_NAME):
            annotations[LEASE_TIME_ANNOTATION_NAME] = current_time.isoformat()

        # If the current leader's lease has timed out than take ownership
        elif not self._check_lease_time(
            annotations.get(LEASE_TIME_ANNOTATION_NAME), current_time
        ):
            annotations[LEASE_NAME_ANNOTATION_NAME] = self.pod_name
            annotations[LEASE_TIME_ANNOTATION_NAME] = current_time.isoformat()

        # Otherwise unable to acquire lock
        else:
            return False

        success, _ = self.deploy_manager.deploy([current_resource])
        if not success:
            log.warning(
                "Unable to update resource annotation%s/%s/%s/%s",
                resource.kind,
                resource.api_version,
                resource.namespace,
                resource.name,
            )
            return False

        return True

    def release(self):
        """
        Release lock on global resource
        """
        return True

    def release_resource(self, resource: ManagedObject):
        """
        Release lock on specific resource by removing the annotation
        """
        current_resource = copy(resource.definition)

        # Only clear annotation if we're the current leader
        if self.pod_name == current_resource["metadata"].get("annotations", {}).get(
            LEASE_NAME_ANNOTATION_NAME
        ):
            current_resource["metadata"]["annotations"][
                LEASE_NAME_ANNOTATION_NAME
            ] = None
            current_resource["metadata"]["annotations"][
                LEASE_TIME_ANNOTATION_NAME
            ] = None
            self.deploy_manager.deploy([current_resource])

        return True

    def is_leader(self, resource: Optional[ManagedObject] = None):
        """
        Determines if current instance is leader
        """
        if resource:
            annotations = resource.metadata.get("annotations", {})
            return self.pod_name == annotations.get(
                LEASE_NAME_ANNOTATION_NAME
            ) and self._check_lease_time(annotations.get(LEASE_TIME_ANNOTATION_NAME))

        return True

    def _check_lease_time(
        self, lease_time: str, current_time: Optional[datetime] = None
    ) -> bool:
        """Helper function to check if lease time is still valid

        Args:
            lease_time: str
                A datetime in isoformat
            current_time: Optional[datetime]
                The time to compare the lease_time to. Use datetime.now() if None

        Returns:
            valid_lease: bool
                If the lease should still be an owner
        """
        # Don't default to datetime.now() in function args as that's only evaluated once
        current_time = current_time or datetime.now()
        return current_time < datetime.fromisoformat(lease_time) + self.duration_delta



"""
Test the DAG CompletionState class
"""

# First Party
import alog

# Local
from oper8.dag import CompletionState, Node

log = alog.use_channel("TEST")

################################################################################
## Completion State #######################################################################
################################################################################


def test_dag_completion_state_str():
    """Coverage test to ensure that stringifying a CompletionState doesn't
    throw! Coverage :)
    """
    log.debug(
        str(
            CompletionState(
                verified_nodes=[Node("A")],
                unverified_nodes=[Node("B")],
                failed_nodes=[Node("C")],
                unstarted_nodes=[Node("D")],
            )
        )
    )



"""
Utils and common classes for the python watch manager tests
"""
# Standard
from datetime import datetime
from multiprocessing.connection import Connection
from queue import Queue
from threading import Event
from uuid import uuid4
import multiprocessing
import random
import tempfile
import time

# Third Party
import pytest

# First Party
import alog

# Local
from oper8.managed_object import ManagedObject
from oper8.reconcile import ReconciliationResult
from oper8.watch_manager.python_watch_manager.filters.common import (
    get_configured_filter,
)
from oper8.watch_manager.python_watch_manager.leader_election import (
    LeadershipManagerBase,
)
from oper8.watch_manager.python_watch_manager.leader_election.lease import (
    LeaderWithLeaseManager,
)
from oper8.watch_manager.python_watch_manager.leader_election.life import (
    LeaderForLifeManager,
)
from oper8.watch_manager.python_watch_manager.threads.heartbeat import HeartbeatThread
from oper8.watch_manager.python_watch_manager.threads.reconcile import ReconcileThread
from oper8.watch_manager.python_watch_manager.threads.timer import TimerThread
from oper8.watch_manager.python_watch_manager.utils.types import (
    ReconcileProcess,
    ReconcileRequest,
    WatchRequest,
)

log = alog.use_channel("TEST")


### Mock Classes
class DisabledLeadershipManager(LeadershipManagerBase):
    """Leadership Manager that is always disabled"""

    def __init__(self):
        self.shutdown_event = Event()

    def acquire_resource(self, resource):
        return False

    def acquire(self, force: bool = False) -> bool:
        if force:
            self.shutdown_event.set()
        return self.shutdown_event.wait()

    def release(self):
        raise NotImplementedError()

    def release_resource(self, resource=None):
        raise NotImplementedError()

    def is_leader(self):
        return False


class MockedLeaderForLifeManager(LeaderForLifeManager):
    _disable_singleton = True


class MockedLeaderWithLeaseManager(LeaderWithLeaseManager):
    _disable_singleton = True


class MockedTimerThread(TimerThread):
    _disable_singleton = True


class MockedHeartbeatThread(HeartbeatThread):
    _disable_singleton = True


class MockedReconcileThread(ReconcileThread):
    """Subclass of ReconcileThread that mocks the subprocess. This was more
    reliable than using unittest.mock"""

    _disable_singleton = True

    def __init__(
        self,
        deploy_manager=None,
        leadership_manager=None,
        subprocess_wait_time=0.1,
        returned_messages=None,
    ):
        self.requests = Queue()
        self.timer_events = Queue()
        self.processes_started = 0
        self.processes_finished = 0
        self.watch_threads_created = 0
        self.subprocess_wait_time = subprocess_wait_time
        self.returned_messages = returned_messages or [[]]
        super().__init__(deploy_manager, leadership_manager)

    def push_request(self, request: ReconcileRequest):
        self.requests.put(request)
        super().push_request(request)

    def get_request(self) -> ReconcileRequest:
        return self.requests.get()

    def _handle_watch_request(self, request: WatchRequest):
        self.watch_threads_created += 1
        return super()._handle_watch_request(request)

    def _handle_process_end(self, reconcile_process: ReconcileProcess):
        self.processes_finished += 1
        return super()._handle_process_end(reconcile_process)

    def _start_reconcile_process(
        self, request: ReconcileRequest, pipe: Connection
    ) -> multiprocessing.Process:
        self.processes_started += 1

        returned_messages = []
        if len(self.returned_messages) > 0:
            returned_messages = self.returned_messages.pop(0)

        # Create and start a mocked reconcile process
        process = self.spawn_ctx.Process(
            target=mocked_create_and_start_entrypoint,
            args=[
                self.logging_queue,
                request,
                pipe,
                self.subprocess_wait_time,
                returned_messages,
            ],
        )
        process.start()
        log.debug3(f"Started child process with pid: {process.pid}")

        return process

    def _create_timer_event_for_request(
        self, request: ReconcileRequest, result: ReconciliationResult = None
    ):
        timer_event = super()._create_timer_event_for_request(request, result)
        self.timer_events.put(timer_event)
        return timer_event


### Helper Fixtures
@pytest.fixture(autouse=True)
def clear_caches():
    get_configured_filter.cache_clear()


### Helper functions
def make_ownerref(resource):
    metadata = resource.get("metadata", {})
    return {
        "apiVersion": resource.get("apiVersion"),
        "kind": resource.get("kind"),
        "name": metadata.get("name"),
        "uid": metadata.get("uid"),
    }


def make_resource(
    kind="Foo",
    namespace="test",
    api_version="foo.bar.com/v1",
    name="foo",
    spec=None,
    status=None,
    generation=1,
    resource_version=None,
    annotations=None,
    labels=None,
    owner_refs=None,
):
    return {
        "kind": kind,
        "apiVersion": api_version,
        "metadata": {
            "name": name,
            "namespace": namespace,
            "generation": generation,
            "resourceVersion": resource_version or random.randint(1, 1000),
            "ownerReferences": owner_refs or [],
            "labels": labels or {},
            "uid": str(uuid4()),
            "annotations": annotations or {},
        },
        "spec": spec or {},
        "status": status or {},
    }


def make_managed_object(*args, **kwargs):
    return ManagedObject(make_resource(*args, **kwargs))


def mocked_create_and_start_entrypoint(
    logging_queue: multiprocessing.Queue,
    request: ReconcileRequest,
    result_pipe: Connection,
    wait_time=0.5,
    returned_messages=None,
):
    """"""
    time.sleep(wait_time)
    for message in returned_messages or []:
        result_pipe.send(message)


def read_heartbeat_file(hb_file: str) -> datetime:
    """Parse a heartbeat file into a datetime"""
    with open(hb_file) as handle:
        hb_str = handle.read()

    return datetime.strptime(hb_str, HeartbeatThread._DATE_FORMAT)


@pytest.fixture
def heartbeat_file():
    with tempfile.NamedTemporaryFile() as tmp_file:
        yield tmp_file.name



# Local
from .connection_base import DatastoreConnectionBase
from .cos import CosFactory
from .factory_base import DatastoreSingletonFactoryBase
from .interfaces import Datastore
from .postgres import PostgresFactory
from .redis import RedisFactory



"""Tests for the VCS class"""

# Standard
from unittest import mock
import pathlib
import shutil
import subprocess
import tempfile

# Third Party
import pygit2
import pytest

# First Party
import alog

# Local
from oper8.test_helpers.helpers import default_branch_name, vcs_project
from oper8.vcs import (
    VCS,
    VCSCheckoutMethod,
    VCSConfigError,
    VCSMultiProcessError,
    VCSRuntimeError,
)

log = alog.use_channel("TEST")


################################################################################
## Helpers #####################################################################
################################################################################


def validate_vcs_dir(directory, refish):
    """Validate git directory using git fsck"""

    # Validate git directory integrity
    subprocess.run(
        [
            "git",
            "-C",
            directory,
            "fsck",
        ],
        check=True,
    )

    # Validate that the checked out commit matches the expected one
    expected_commit_hash = subprocess.run(
        ["git", "-C", directory, "rev-parse", refish], capture_output=True
    ).stdout
    current_commit_hash = subprocess.run(
        ["git", "-C", directory, "rev-parse", "HEAD"], capture_output=True
    ).stdout
    assert current_commit_hash == expected_commit_hash


################################################################################
## Tests #######################################################################
################################################################################

##################
## Construction ##
##################


def test_construct_defaults(vcs_project):
    """Make sure that a VCS can be constructed with a generic repository"""
    vcs = VCS(vcs_project)
    assert vcs.repo


def test_construct_defaults_invalid_repo():
    """Make sure that a VCS can be constructed with a generic repository"""
    with tempfile.TemporaryDirectory() as temp_dir:
        with pytest.raises(VCSConfigError):
            vcs = VCS(temp_dir)


##################
## get_ref ##
##################


@pytest.mark.parametrize(
    ["refish", "exception"],
    [
        ["1.2.3", None],
        [default_branch_name(), None],
        ["refs/tags/1.2.3", None],
        ["invalid", VCSConfigError],
    ],
)
def test_get_ref(vcs_project, refish, exception):
    vcs = VCS(vcs_project)

    if exception:
        with pytest.raises(exception):
            vcs.get_ref(refish)
    else:
        vcs.get_ref(refish)


##################
## list_refs ##
##################


def test_list_refs(vcs_project):
    vcs = VCS(vcs_project)

    assert "1.2.3" in vcs.list_refs()
    assert default_branch_name() in vcs.list_refs()
    # Ensure only short hand versions are listed
    assert "refs/tags/1.2.3" not in vcs.list_refs()


##################
## checkout_ref ##
##################


@pytest.mark.parametrize(
    ["refish", "method", "exception"],
    [
        ["1.2.3", VCSCheckoutMethod.WORKTREE, None],
        ["1.2.3", VCSCheckoutMethod.CLONE, None],
        ["1.2.3", "invalid", VCSConfigError],
        ["wrong", VCSCheckoutMethod.CLONE, VCSConfigError],
    ],
)
def test_checkout_ref(vcs_project, refish, method, exception):
    with tempfile.TemporaryDirectory() as dest_dir:
        checkout_path = pathlib.Path(dest_dir) / "dest_repo"

        vcs = VCS(vcs_project)

        if exception:
            with pytest.raises(exception):
                vcs.checkout_ref(dest_path=checkout_path, refish=refish, method=method)
        else:
            vcs.checkout_ref(dest_path=checkout_path, refish=refish, method=method)
            validate_vcs_dir(checkout_path, refish)


def test_checkout_ref_existing(vcs_project):
    with tempfile.TemporaryDirectory() as dest_dir:
        dest_path = pathlib.Path(dest_dir)

        vcs = VCS(vcs_project)

        # Copy the vcs project to the destination manually
        shutil.copytree(vcs_project, dest_path, dirs_exist_ok=True)

        # Validate that checking out the same commit to the same destination
        # works as expected
        vcs.checkout_ref(
            dest_path=dest_path,
            refish=default_branch_name(),
        )

        # Check that two refishs that point to the same commit are able to reuse
        # directory
        with pytest.raises(VCSConfigError):
            vcs.checkout_ref(
                dest_path=dest_path,
                refish="1.2.3",
            )

        # Ensure multiprocessing error if destination git directory isn't fully checked out
        (dest_path / ".git" / "index").unlink()
        with pytest.raises(VCSMultiProcessError):
            vcs.checkout_ref(
                dest_path=dest_path,
                refish=default_branch_name(),
            )


def test_clone_ref(vcs_project):
    with tempfile.TemporaryDirectory() as dest_dir:
        dest_path = pathlib.Path(dest_dir)
        vcs = VCS(vcs_project)

        _, ref = vcs.get_ref(default_branch_name())

        vcs._clone_ref(
            dest_path=dest_path,
            ref=ref,
        )

        # Ensure cloned a valid directory
        validate_vcs_dir(dest_path, default_branch_name())
        # Ensure the entire git directory was copied
        assert (dest_path / ".git").is_dir()


def test_clone_ref_exceptions(vcs_project):
    vcs = VCS(vcs_project)
    _, ref = vcs.get_ref("")

    with pytest.raises(VCSRuntimeError):
        vcs._clone_ref(
            dest_path=None,
            ref=ref,
        )


def test_create_branch(vcs_project):
    vcs = VCS(vcs_project)

    commit, _ = vcs.get_ref(default_branch_name())

    # Ensure branch creation works
    vcs.create_branch(
        branch_name="test_branch",
        commit=commit,
    )

    # Ensure branch creation works if an existing branch is already created
    vcs.create_branch(branch_name="test_branch", commit=commit)


def test_create_branch_exceptions(vcs_project):
    vcs = VCS(vcs_project)

    # Test creating a branch with a same name but different commit fails
    commit, _ = vcs.get_ref("1.2.3")
    with pytest.raises(VCSRuntimeError):
        vcs.create_branch(branch_name=default_branch_name(), commit=commit)

    # We can't easily simulate branch multiprocessing errors so mock them
    vcs.repo.branches.create = mock.Mock(side_effect=pygit2.AlreadyExistsError)
    with pytest.raises(VCSMultiProcessError):
        vcs.create_branch(branch_name="dev_branch", commit=commit)

    vcs.repo.branches.create.side_effect = OSError
    with pytest.raises(VCSRuntimeError):
        vcs.create_branch(branch_name="dev_branch", commit=commit)


def test_create_worktree(vcs_project):
    with tempfile.TemporaryDirectory() as dest_dir:
        dest_path = pathlib.Path(dest_dir) / "dest_repo"
        vcs = VCS(vcs_project)

        vcs._create_worktree("test", dest_path, branch=vcs.repo.branches.get("1.2.3"))

        validate_vcs_dir(dest_path, "1.2.3")


def test_create_worktree_exceptions(vcs_project):
    with tempfile.TemporaryDirectory() as dest_dir:
        dest_path = pathlib.Path(dest_dir)
        vcs = VCS(vcs_project)

        # If we try to create a worktree using an existing directory
        with pytest.raises(VCSMultiProcessError):
            vcs._create_worktree(
                "test", dest_path, branch=vcs.repo.branches.get("1.2.3")
            )

        # If we try to create a worktree for a branch that is already checked out
        # then raise a multiprocessing error
        with pytest.raises(VCSMultiProcessError):
            vcs._create_worktree(
                "test",
                dest_path / "dest_dir",
                branch=vcs.repo.branches.get(default_branch_name()),
            )

        # Simulate some other vcs error
        vcs.repo.add_worktree = mock.Mock(side_effect=pygit2.GitError)
        with pytest.raises(VCSRuntimeError):
            vcs._create_worktree(
                "test",
                dest_path / "dest_dir",
                branch=vcs.repo.branches.get(default_branch_name()),
            )



"""
Tests for the OpenshiftDeployManager
"""

# Standard
from contextlib import contextmanager
from queue import Queue
from threading import Event, Thread
from unittest import mock
import time

# Third Party
import pytest

# First Party
import alog

# Local
from oper8 import config
from oper8 import status as oper8_status
from oper8.deploy_manager.base import DeployMethod
from oper8.deploy_manager.kube_event import KubeEventType
from oper8.deploy_manager.openshift_deploy_manager import OpenshiftDeployManager
from oper8.deploy_manager.owner_references import _make_owner_reference
from oper8.test_helpers.helpers import (
    SOME_OTHER_NAMESPACE,
    TEST_NAMESPACE,
    MockedOpenshiftDeployManager,
    configure_logging,
    library_config,
)
from oper8.test_helpers.kub_mock import MockKubClient, mock_kub_client_constructor
from oper8.utils import merge_configs

## Helpers #####################################################################

configure_logging()
log = alog.use_channel("TEST")


def setup_testable_manager(manage_ansible_status=False, owner_cr=None, *args, **kwargs):
    """Set up a testable deploy manager with a mock api client"""
    dm = MockedOpenshiftDeployManager(
        manage_ansible_status=manage_ansible_status,
        owner_cr=owner_cr,
        *args,
        **kwargs,
    )
    dm.client.resources.invalidate_cache()
    return dm


def make_obj_states(cluster_state, name_override=None):
    """Make a list of object state dicts for each object in the cluster state"""
    states = []
    for namespace, ns_entries in cluster_state.items():
        for kind, kind_entries in ns_entries.items():
            for api_version, version_entries in kind_entries.items():
                for name, state in version_entries.items():
                    if callable(state):
                        state = state("GET", namespace, kind, api_version, name)
                    if isinstance(state, tuple):
                        state = state[0]
                    states.append(
                        merge_configs(
                            state,
                            {
                                "apiVersion": api_version,
                                "kind": kind,
                                "metadata": {
                                    "name": name_override or name,
                                    "namespace": namespace,
                                },
                            },
                        )
                    )
    return states


def get_ansible_conditions(status):
    return [
        cond
        for cond in status.get("conditions", [])
        if cond.get("type") == OpenshiftDeployManager._ANSIBLE_COND_TYPE
    ]


@contextmanager
def gather_watch_events(dm, *args, expected_events=None, **kwargs):
    """Helper function to pipe watch events from a separate thread into a queue"""

    # Wait until the watch has started before returning. This prevents
    # duplicate events
    startup = Event()
    add_watch_queue = dm.client.client._add_watch_queue

    def startup_watch(*args, **kwargs):
        log.debug2("Setting startup event")
        startup.set()
        return add_watch_queue(*args, **kwargs)

    dm.client.client._add_watch_queue = mock.Mock(side_effect=startup_watch)

    # Setup watch queue for consuming events
    watch_queue = Queue()

    def watch_events():
        log.debug3("Watch Thread started")
        num_events = 0
        for event in dm.watch_objects(*args, **kwargs):
            log.debug2("Received event %s", event)
            watch_queue.put(event)

            # if we've gotten the expected number of events
            # exit
            num_events += 1
            if num_events >= expected_events:
                log.debug2("Exiting watch thread")
                return

    #  Start a thread for watching events in the background
    watch_thread = Thread(target=watch_events, daemon=True)
    watch_thread.start()

    # Wait for watch before starting
    log.debug2("Waiting for watch to start")
    while not startup.is_set():
        pass

    log.debug2("Watch started")
    yield watch_queue

    # If the number of expected events was given assert that the thread exited
    if expected_events:
        watch_thread.join(1)
        assert not watch_thread.is_alive(), "Extra events in watch queue"


## Tests #######################################################################


############
## deploy ##
############


def test_deploy_new_resource():
    """Make sure that deploying a new instance of an existing resource type to
    the cluster works
    """
    start_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {}}}}
    end_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    dm = setup_testable_manager(cluster_state=start_cluster_state)
    success, changed = dm.deploy(make_obj_states(end_cluster_state))
    assert success
    assert changed


def test_deploy_update_resource_change():
    """Make sure that deploying a change to an existing resource instance in the
    cluster works when there is a valid change
    """
    start_cluster_state = {
        "test": {"Foo": {"foo.bar.com/v1": {"bar": {"wing": "bat"}}}}
    }
    end_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {"wing": "bar"}}}}}
    dm = setup_testable_manager(cluster_state=start_cluster_state)

    # Deploy it and make sure it changes
    success, changed = dm.deploy(make_obj_states(end_cluster_state))
    assert success
    assert changed

    # Look it back up and make sure the changes took
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success
    assert content is not None
    assert content.get("wing") == "bar"


def test_deploy_update_resource_no_change():
    """Make sure that deploying a change to an existing resource instance in the
    cluster works when there is no change
    """
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {"wing": "bat"}}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)
    success, changed = dm.deploy(make_obj_states(cluster_state))
    assert success
    assert not changed


def test_deploy_method_resource():
    """Make sure that deploying a new instance of an existing resource type to
    the cluster works
    """
    start_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {}}}}
    end_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    replace_apply_resource = {
        "kind": "Foo",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {"name": "bar", "namespace": "test"},
        "spec": {"some": "key_1"},
    }
    end_apply_resource = {
        "kind": "Foo",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {"name": "bar", "namespace": "test"},
        "spec": {"some": "key_2"},
    }
    dm = setup_testable_manager(cluster_state=start_cluster_state)
    dm._requires_replace = lambda *args, **kwargs: True

    replace_called = []

    def track_replace(resource_definition: dict):
        replace_called.append(resource_definition)
        return MockedOpenshiftDeployManager._replace_resource(dm, resource_definition)

    dm._replace_resource = track_replace

    apply_called = []

    def track_apply(resource_definition):
        apply_called.append(resource_definition)
        return MockedOpenshiftDeployManager._apply_resource(dm, resource_definition)

    dm._apply_resource = track_apply

    # Use apply instead of replace when first deploying
    success, changed = dm.deploy(
        resource_definitions=make_obj_states(end_cluster_state),
        method=DeployMethod.REPLACE,
    )
    assert success
    assert changed
    assert len(replace_called) == 0 and len(apply_called) == 1

    success, changed = dm.deploy(
        resource_definitions=[replace_apply_resource],
        method=DeployMethod.REPLACE,
    )
    assert success
    assert changed
    assert len(replace_called) == 1 and len(apply_called) == 1

    success, changed = dm.deploy(
        resource_definitions=[end_apply_resource], method=DeployMethod.UPDATE
    )
    assert success
    assert changed
    assert len(replace_called) == 1 and len(apply_called) == 2


def test_deploy_method_update_resource():
    """Make sure that deploying a new instance of an existing resource type to
    the cluster works
    """
    start_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {}}}}
    end_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    dm = setup_testable_manager(cluster_state=start_cluster_state)
    success, changed = dm.deploy(
        make_obj_states(end_cluster_state), DeployMethod.UPDATE
    )
    assert success
    assert changed


def test_deploy_unknown_resource_type():
    """Make sure that trying to deploy an instance of an unknown resource type
    results in a handled error
    """
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {"wing": "bat"}}}}}
    dm = setup_testable_manager(cluster_state={})
    success, changed = dm.deploy(make_obj_states(cluster_state))
    assert not success
    assert not changed


def test_deploy_conflict_persist():
    """Make sure that trying to deploy and encountering a persistent 409
    conflict error results in a failed deployment
    """
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": ({}, 409)}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)
    success, changed = dm.deploy(make_obj_states(cluster_state))
    assert not success
    assert not changed


def test_deploy_conflict_retry_resolve_count():
    """Make sure that trying to deploy and encountering a 409 conflict error the
    first time which resolves on retry results in a successful deployment
    """

    class ConflictOncePatch:
        def __init__(self):
            self._count = 0

        def __call__(self, method, namespace, kind, api_version, name):
            log.debug(
                "Handling call with method [%s] for [%s/%s/%s/%s]",
                method,
                namespace,
                api_version,
                kind,
                name,
            )
            res = {
                "metadata": {
                    "resourceVersion": self._count,
                }
            }
            if method == "PATCH":
                if self._count < 1:
                    log.debug("Returning 409")
                    res = ({}, 409)
                self._count += 1
            return res

    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": ConflictOncePatch()}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {"some": "content"}}}}}
    success, changed = dm.deploy(make_obj_states(cluster_state))
    assert success
    assert changed


@pytest.mark.parametrize("enable_fallback", [True, False])
def test_deploy_conflict_unprocessible_retry_resolve_count(enable_fallback):
    """Make sure that trying to deploy and encountering a 409 conflict error the
    first time, then encountering a 422 falls back to using PUT and resolves
    """

    class ConflictUnprocessiblePatch:
        def __init__(self):
            self._count = 0

        def __call__(self, method, namespace, kind, api_version, name):
            log.debug(
                "Handling call with method [%s] for [%s/%s/%s/%s]",
                method,
                namespace,
                api_version,
                kind,
                name,
            )
            res = {
                "metadata": {
                    "resourceVersion": self._count,
                }
            }
            if method == "PATCH":
                if self._count < 1:
                    log.debug("Returning 409")
                    res = ({}, 409)
                else:
                    log.debug("Returning 422")
                    res = ({}, 422)
                self._count += 1
            return res

    with library_config(deploy_unprocessable_put_fallback=enable_fallback):
        cluster_state = {
            "test": {"Foo": {"foo.bar.com/v1": {"bar": ConflictUnprocessiblePatch()}}}
        }
        dm = setup_testable_manager(cluster_state=cluster_state)
        cluster_state = {
            "test": {"Foo": {"foo.bar.com/v1": {"bar": {"some": "content"}}}}
        }
        assert dm.deploy(make_obj_states(cluster_state))[0] is enable_fallback


def test_deploy_conflict_retry_resolve_time():
    """Make sure that trying to deploy and encountering a 409 conflict error for
    a period of time eventually results in a successful deploy
    """

    class ConflictTimePatch:
        def __init__(self):
            self._first = None

        def __call__(self, method, namespace, kind, api_version, name):
            log.debug(
                "Handling call with method [%s] for [%s/%s/%s/%s]",
                method,
                namespace,
                api_version,
                kind,
                name,
            )
            res = {
                "metadata": {
                    "resourceVersion": 1234,
                }
            }
            if method == "PATCH":
                if self._first is None:
                    self._first = time.time()
                if time.time() - self._first < config.retry_backoff_base_seconds * 2.5:
                    log.debug("Returning 409")
                    res = ({}, 409)
            return res

    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": ConflictTimePatch()}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {"some": "content"}}}}}
    success, changed = dm.deploy(make_obj_states(cluster_state))
    assert success
    assert changed


def test_deploy_forbidden():
    """Make sure that a forbidden error when deploying results in a handled
    error
    """
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": ({}, 403)}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)
    success, changed = dm.deploy(make_obj_states(cluster_state))
    assert not success
    assert not changed


def test_deploy_no_cached():
    """Make sure that when an object is deployed, a subsequent call to
    get_object_current_state will pull from the cluster to get the updated state
    """
    start_cluster_state = {
        "test": {"Foo": {"foo.bar.com/v1": {"bar": {"wing": "bat"}}}}
    }
    end_cluster_state = {
        "test": {"Foo": {"foo.bar.com/v1": {"bar": {"wom": "bar", "wing": "bar"}}}}
    }
    dm = setup_testable_manager(cluster_state=start_cluster_state)

    # Fetch it to pre-populate the cache
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success
    assert content is not None
    assert "wom" not in content
    assert content.get("wing") == "bat"

    # Deploy it and make sure it changes
    success, changed = dm.deploy(make_obj_states(end_cluster_state))
    assert success
    assert changed

    # Look it back up and make sure the changes took
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success
    assert content is not None
    assert content.get("wom") == "bar"
    assert content.get("wing") == "bar"


def test_deploy_multiple_resources_ok():
    """Make sure that deploying multiple valid resources at a time succeeds"""
    start_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {}}}}
    end_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}, "baz": {}}}}}
    dm = setup_testable_manager(cluster_state=start_cluster_state)
    success, changed = dm.deploy(make_obj_states(end_cluster_state))
    assert success and changed
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success and content is not None
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="baz"
    )
    assert success and content is not None


def test_deploy_multiple_resources_second_bad_content():
    """Make sure that deploying multiple resources where the first is valid and
    the second is malformed successfully applies the first one, but fails on the
    second one
    """

    def patch_unprocessable(method, namespace, kind, api_version, name):
        if method == "PATCH":
            return ({}, 422)
        else:
            return {}

    start_cluster_state = {
        "test": {"Foo": {"foo.bar.com/v1": {"baz": patch_unprocessable}}}
    }
    deploy_cluster_state = {
        "test": {
            "Foo": {
                "foo.bar.com/v1": {
                    "bar": {},
                    "baz": {"something": "different"},
                }
            }
        }
    }
    dm = setup_testable_manager(cluster_state=start_cluster_state)
    success, changed = dm.deploy(make_obj_states(deploy_cluster_state))
    assert not success
    assert changed

    # Make sure "bar" actually got deployed
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success and content is not None


def test_deploy_multiple_resources_first_bad_content():
    """Make sure that deploying multiple resources where the first is invalid
    and the second would be valid does not deploy the second after the first
    fails
    """

    def patch_unprocessable(method, namespace, kind, api_version, name):
        if method == "PATCH":
            return ({}, 422)
        else:
            return {}

    start_cluster_state = {
        "test": {"Foo": {"foo.bar.com/v1": {"bar": patch_unprocessable}}}
    }
    deploy_cluster_state = {
        "test": {
            "Foo": {
                "foo.bar.com/v1": {
                    "bar": {"something": "different"},
                    "baz": {},
                }
            }
        }
    }
    dm = setup_testable_manager(cluster_state=start_cluster_state)
    success, changed = dm.deploy(make_obj_states(deploy_cluster_state))
    assert not success
    assert not changed

    # Make sure "baz" did not get deployed
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="baz"
    )
    assert success and content is None


def test_deploy_empty_list():
    """Make sure that an empty list of resources is a no-op"""
    dm = setup_testable_manager()
    success, changed = dm.deploy([])
    assert success
    assert not changed


def test_deploy_add_owner_reference():
    """Make sure that deploying a resource with an owner CR adds the owner
    reference given the resource and owner CR are in the same namespace
    """
    start_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {}}}}
    end_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    owner_cr = {
        "kind": "Owner",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {"name": "owner", "namespace": "test", "uid": "unique"},
    }
    dm = setup_testable_manager(cluster_state=start_cluster_state, owner_cr=owner_cr)
    success, changed = dm.deploy(make_obj_states(end_cluster_state))
    assert success
    assert changed
    success, content = dm.get_object_current_state(
        kind="Foo",
        api_version="foo.bar.com/v1",
        namespace="test",
        name="bar",
    )
    assert success
    assert content["metadata"]["ownerReferences"] == [_make_owner_reference(owner_cr)]


def test_deploy_dont_add_owner_reference_if_manage_owner_references_disabled():
    """Make sure that deploying a resource with an owner CR does not add the
    owner reference when manage_owner_references is False
    """
    start_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {}}}}
    end_cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    owner_cr = {
        "kind": "Owner",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {"name": "owner", "namespace": "test", "uid": "unique"},
    }
    dm = setup_testable_manager(cluster_state=start_cluster_state, owner_cr=owner_cr)
    success, changed = dm.deploy(
        make_obj_states(end_cluster_state), manage_owner_references=False
    )
    assert success
    assert changed
    success, content = dm.get_object_current_state(
        kind="Foo",
        api_version="foo.bar.com/v1",
        namespace="test",
        name="bar",
    )
    assert success
    assert "ownerReferences" not in content["metadata"]


def test_deploy_cluster_wide_resource():
    """Make sure that deploying a new instance of an existing resource type to
    the cluster works
    """
    start_cluster_state = {"": {"Foo": {"foo.bar.com/v1": {}}}}
    dm = setup_testable_manager(cluster_state=start_cluster_state)

    obj = {"apiVersion": "foo.bar.com/v1", "kind": "Foo", "metadata": {"name": "bar"}}
    success, changed = dm.deploy([obj])

    assert success
    assert changed


def test_deploy_dont_add_owner_reference_if_resource_and_owner_cr_not_in_same_namespace():
    """Make sure that deploying a resource with an owner CR does not add the owner
    reference if the resource and owner CR are in a different namespace
    """
    start_cluster_state = {SOME_OTHER_NAMESPACE: {"Foo": {"foo.bar.com/v1": {}}}}
    end_cluster_state = {SOME_OTHER_NAMESPACE: {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    owner_cr = {
        "kind": "Owner",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {"name": "owner", "namespace": TEST_NAMESPACE, "uid": "unique"},
    }
    dm = setup_testable_manager(cluster_state=start_cluster_state, owner_cr=owner_cr)
    success, changed = dm.deploy(make_obj_states(end_cluster_state))
    assert success
    assert changed
    success, content = dm.get_object_current_state(
        kind="Foo",
        api_version="foo.bar.com/v1",
        namespace=SOME_OTHER_NAMESPACE,
        name="bar",
    )
    assert success
    assert content["metadata"]["ownerReferences"] == []


#############
## disable ##
#############


def test_disable_remove_present_resource():
    """Make sure that disabling a resource which is in the cluster correctly
    deletes it
    """
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)
    success, changed = dm.disable(make_obj_states(cluster_state))
    assert success
    assert changed


def test_disable_remove_object_not_present():
    """Make sure that disabling a resource which is in the cluster correctly
    indicates that no change was made
    """
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)
    success, changed = dm.disable(make_obj_states(cluster_state, name_override="baz"))
    assert success
    assert not changed


def test_disable_remove_kind_not_present():
    """Make sure that disabling a kind which is unknown to the cluster indicates
    success (the resource doesn't exist at the end)
    """
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    dm = setup_testable_manager()
    success, changed = dm.disable(make_obj_states(cluster_state))
    assert success
    assert not changed


def test_disable_remove_object_non_namespaced():
    """Make sure that disabling a resource which is in the cluster correctly
    indicates that no change was made
    """
    cluster_state = {None: {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)

    obj = make_obj_states(cluster_state)[0]
    del obj["metadata"]["namespace"]
    success, changed = dm.disable([obj])
    assert success
    assert not changed


def test_disable_remove_forbidden():
    """Make sure that a forbidden error is handled when attempting to remove a
    resource from the cluster
    """

    def delete_forbidden(method, namespace, kind, api_version, name):
        if method == "DELETE":
            return ({}, 403)
        else:
            return {}

    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": delete_forbidden}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)
    success, changed = dm.disable(make_obj_states(cluster_state))
    assert not success
    assert not changed


def test_disable_invalid_resource_list():
    """Make sure that passing an invalid type shows non-success"""
    dm = setup_testable_manager()
    with pytest.raises(AssertionError):
        dm.disable("not a list")


def test_disable_clear_cache():
    """Make sure that for an object that has been cached, a disable will cause
    it to not hit cache on a subsequent get_object_current_state
    """
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    dm = setup_testable_manager(cluster_state=cluster_state)

    # First get the object to populate the cache
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success
    assert content is not None

    # Disable it
    success, changed = dm.disable(make_obj_states(cluster_state))
    assert success
    assert changed

    # Look it up again and make sure it's not found
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success
    assert content is None


##############################
## get_object_current_state ##
##############################


def test_get_object_current_state_object_present():
    """Make sure that get_object_current_state returns success and a valid
    object when the object is present in the cluster
    """
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success
    assert content is not None


def test_get_object_current_state_resource_not_present():
    """Make sure that get_object_current_state returns success and None when the
    resource type is not present in the cluster
    """
    dm = setup_testable_manager(cluster_state={})
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar"
    )
    assert success
    assert content is None


def test_get_object_current_state_object_not_present():
    """Make sure that get_object_current_state returns success and None when the
    object is not present in the cluster
    """
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="baz"
    )
    assert success
    assert content is None


def test_get_object_current_state_object_not_unique():
    """Make sure that when a given resource type is given which has multiple
    api_versions available and the api_version is not given as an arg, the
    ResourceNotUniqueError is handled as if the object is not found.
    """
    dm = setup_testable_manager(
        cluster_state={
            "test": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {}},
                    "foo.bar.com/v1beta1": {"bar": {}},
                }
            }
        }
    )
    success, content = dm.get_object_current_state(
        kind="Foo",
        namespace="test",
        name="bar",
    )
    assert success
    assert content is None


def test_get_object_current_state_object_not_unique_with_version():
    """Make sure that when a given resource type is given which has multiple
    api_versions available and the api_version is given as an arg, the
    object is found and returned.
    """
    dm = setup_testable_manager(
        cluster_state={
            "test": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {}},
                    "foo.bar.com/v1beta1": {"bar": {}},
                }
            }
        }
    )
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar", api_version="v1"
    )
    assert success
    assert content is not None


def test_multiple_version_error():
    """Make sure that deploying a new instance of an existing resource type to
    the cluster works
    """
    dm = setup_testable_manager(
        cluster_state={
            "default": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {}},
                    "foo.bar.com/v1beta1": {"bar": {}},
                }
            }
        }
    )

    # Assert no object is returned
    success, obj = dm.get_object_current_state(
        kind="Foo",
        namespace="",
        api_version="",
        name="",
    )

    assert success
    assert not obj

    success, obj = dm.get_object_current_state(
        kind="Foo",
        namespace="default",
        api_version="foo.bar.com/v1",
        name="bar",
    )

    assert success
    assert obj


def test_get_object_current_state_forbidden():
    """Make sure that if a 403 (forbidden) is given, it's reported as
    non-successful
    """
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": ({}, 403)}}}}
    )
    success, content = dm.get_object_current_state(
        kind="Foo", namespace="test", name="bar", api_version="v1"
    )
    assert not success
    assert content is None


def test_get_object_current_state_object_not_namespaced():
    """Make sure that when a given resource type is given which has multiple
    api_versions available and the api_version is not given as an arg, the
    ResourceNotUniqueError is handled as if the object is not found.
    """
    dm = setup_testable_manager(
        cluster_state={
            "": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {}},
                    "foo.bar.com/v1beta1": {"bar": {}},
                }
            }
        }
    )
    success, content = dm.get_object_current_state(
        kind="Foo", name="bar", api_version="v1beta1"
    )
    assert success
    assert content


##################################
## filter_objects_current_state ##
##################################


def test_filter_objects_current_state_object_present():
    """Make sure that filter_objects_current_state returns success and a valid
    object when the object is present in the cluster
    """
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )
    success, content = dm.filter_objects_current_state(kind="Foo", namespace="test")
    assert success
    assert len(content) != 0


def test_filter_objects_current_state_resource_not_present():
    """Make sure that filter_objects_current_state returns success and [] when the
    resource type is not present in the cluster
    """
    dm = setup_testable_manager(cluster_state={})
    success, content = dm.filter_objects_current_state(kind="Foo", namespace="test")
    assert success
    assert len(content) == 0


def test_filter_objects_current_state_object_label_selector():
    """Make sure that filter_objects_current_state returns success and [obj] when an
    object matches the label selector
    """
    dm = setup_testable_manager(
        cluster_state={
            "test": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {"metadata": {"labels": {"app": "bar"}}}}
                }
            }
        }
    )
    success, content = dm.filter_objects_current_state(
        kind="Foo", namespace="test", label_selector="app=bar"
    )
    assert success
    assert len(content) == 1


def test_filter_objects_current_state_object_field_selector():
    """Make sure that filter_objects_current_state returns success and [obj] when an
    object matches the field selector
    """
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {"spec": {"a": 1}}}}}}
    )
    success, content = dm.filter_objects_current_state(
        kind="Foo", namespace="test", field_selector="spec.a==1"
    )
    assert success
    assert len(content) == 1


def test_filter_objects_current_state_object_not_unique():
    """Make sure that when a given resource type is given which has multiple
    api_versions available and the api_version is not given as an arg, the
    ResourceNotUniqueError is handled as if the object is not found.
    """
    dm = setup_testable_manager(
        cluster_state={
            "test": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {}},
                    "foo.bar.com/v1beta1": {"bar": {}},
                }
            }
        }
    )
    success, content = dm.filter_objects_current_state(kind="Foo", namespace="test")
    assert success
    assert len(content) == 0


def test_filter_objects_current_state_object_not_unique_with_version():
    """Make sure that when a given resource type is given which has multiple
    api_versions available and the api_version is given as an arg, the
    object is found and returned.
    """
    dm = setup_testable_manager(
        cluster_state={
            "test": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {}},
                    "foo.bar.com/v1beta1": {"bar": {}},
                }
            }
        }
    )
    success, content = dm.filter_objects_current_state(
        kind="Foo", namespace="test", api_version="v1"
    )
    assert success
    assert len(content) != 0


def test_multiple_version_error():
    """Make sure that deploying a new instance of an existing resource type to
    the cluster works
    """
    dm = setup_testable_manager(
        cluster_state={
            "default": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {}},
                    "foo.bar.com/v1beta1": {"bar": {}},
                }
            }
        }
    )

    # Assert no object is returned
    success, obj = dm.filter_objects_current_state(
        kind="Foo",
        namespace="",
        api_version="",
    )

    assert success
    assert len(obj) == 0

    success, obj = dm.filter_objects_current_state(
        kind="Foo",
        namespace="default",
        api_version="foo.bar.com/v1",
    )

    assert success
    assert len(obj) == 1


def test_filter_objects_current_state_forbidden():
    """Make sure that if a 403 (forbidden) is given, it's reported as
    non-successful
    """
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": ({}, 403)}}}}
    )
    success, content = dm.filter_objects_current_state(
        kind="Foo", namespace="test", api_version="v1"
    )
    assert not success
    assert len(content) == 0


def test_filter_objects_current_state_object_not_namespaced():
    """Make sure that when a given resource type is given which has multiple
    api_versions available and the api_version is not given as an arg, the
    ResourceNotUniqueError is handled as if the object is not found.
    """
    dm = setup_testable_manager(
        cluster_state={
            "": {
                "Foo": {
                    "foo.bar.com/v1": {"bar": {}},
                    "foo.bar.com/v1beta1": {"bar": {}},
                }
            }
        }
    )
    success, content = dm.filter_objects_current_state(
        kind="Foo",
        api_version="v1beta1",
    )
    assert success
    assert len(content) != 0


###################
## watch_objects ##
###################


@pytest.mark.timeout(5)
def test_watch_objects():
    dm = setup_testable_manager(cluster_state={"test": {"Foo": {"foo.bar.com/v1": {}}}})
    initial_objects = make_obj_states(
        {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )

    with gather_watch_events(
        dm,
        expected_events=3,
        kind="Foo",
        api_version="foo.bar.com/v1",
        namespace="test",
    ) as watch_queue:

        # Create, Update, and delete resource
        dm.deploy(initial_objects)
        dm.deploy(
            make_obj_states(
                {"test": {"Foo": {"foo.bar.com/v1": {"bar": {"spec": "updated"}}}}}
            )
        )
        dm.disable(initial_objects)

        # Assert each event is captured
        assert watch_queue.get().type == KubeEventType.ADDED
        assert watch_queue.get().type == KubeEventType.MODIFIED
        assert watch_queue.get().type == KubeEventType.DELETED


@pytest.mark.timeout(5)
def test_watch_objects_resourced():
    dm = setup_testable_manager(cluster_state={"test": {"Foo": {"foo.bar.com/v1": {}}}})
    initial_objects = make_obj_states(
        {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )
    dm.deploy(initial_objects)

    with gather_watch_events(
        dm,
        expected_events=2,
        kind="Foo",
        api_version="foo.bar.com/v1",
        namespace="test",
        name="bar",
    ) as watch_queue:

        # Create and Delete objects while updating a non-watched resource. Ensure only
        # the create and delete events are captured
        dm.deploy(
            make_obj_states(
                {
                    "test": {
                        "Foo": {"foo.bar.com/v1": {"different": {"spec": "updated"}}}
                    }
                }
            )
        )
        dm.disable(initial_objects)
        # Assert only events related to our obj get returned
        assert watch_queue.get().type == KubeEventType.ADDED
        assert watch_queue.get().type == KubeEventType.DELETED


################
## set_status ##
################


def test_set_status_happy_path():
    """Make sure that setting a simple status can round-trip safely"""
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )
    log.debug("Setting Status")
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status={"foo": "bar"},
    )
    assert success
    assert changed


def test_set_status_conflict_retry_success():
    """Make sure that if a conflict occurs on the first update, a retry is
    attempted
    """

    # Small callable that will return a conflict the first time, then return
    # success
    class OnePutConflict:
        def __init__(self):
            self.first_call = True

        def __call__(self, method, namespace, kind, api_version, name):
            if method == "PUT" and self.first_call:
                log.debug("First call. Returning 409")
                self.first_call = False
                return ({}, 409)
            else:
                return {"metadata": {"resourceVersion": 123}}

    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": OnePutConflict()}}}}
    )
    log.debug("Setting Status")
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status={"foo": "bar"},
    )
    assert success
    assert changed


def test_set_status_conflict_retry_failure():
    """Make sure that if a conflict occurs on a retry, we do not retry a third
    time
    """

    # Small callable that will return a conflict the first time, then return
    # success
    class TwoPutConflicts:
        def __init__(self):
            self.put_call_num = 0

        def __call__(self, method, namespace, kind, api_version, name):
            if method == "PUT":
                if self.put_call_num < 2:
                    log.debug("Call %d. Returning 409", self.put_call_num)
                    return ({}, 409)
                self.put_call_num += 1
            return {}

    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": TwoPutConflicts()}}}}
    )
    log.debug("Setting Status")
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status={"foo": "bar"},
    )
    not success
    not changed


def test_set_status_update_cache():
    """Make sure that when set_status is called, a subsequent
    get_object_current_state on the same object will return the updated status
    """
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )

    # Fetch current state and make sure status is not there
    success, content = dm.get_object_current_state(
        kind="Foo", name="bar", namespace="test"
    )
    assert success
    assert content is not None
    assert "status" not in content

    # Set the status
    log.debug("Setting Status")
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status={"foo": "bar"},
    )
    assert success
    assert changed

    # Re-fetch and make sure the status is present
    success, content = dm.get_object_current_state(
        kind="Foo", name="bar", namespace="test"
    )
    assert success
    assert content is not None
    assert "status" in content
    assert content["status"].get("foo") == "bar"


def test_set_status_update_no_change():
    """Make sure that when set_status is called, if the new status is the same
    as the current status, the return indicates no change
    """
    status = {"foo": "bar"}
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {"status": status}}}}}
    )
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status=status,
    )
    assert success
    assert not changed


def test_set_status_resource_kind_missing():
    """Make sure that when set_status is called and the resource kind does not
    exist, the function handles it and returns non-success
    """
    dm = setup_testable_manager(cluster_state={})
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status={"foo": "bar"},
    )
    assert not success
    assert not changed


def test_set_status_resource_object_missing():
    """Make sure that when set_status is called and the named object does not
    exist, the function handles it and returns non-success
    """
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )
    success, changed = dm.set_status(
        kind="Foo",
        name="baz",
        namespace="test",
        api_version="foo.bar.com/v1",
        status={"foo": "bar"},
    )
    assert not success
    assert not changed


def test_set_status_put_forbidden():
    """Make sure that when set_status is called and the PUT method returns a
    403, the error is handled and returns non-success
    """

    # Small callable that will return 403 only on PUT
    def put_forbidden(method, namespace, kind, api_version, name):
        if method == "PUT":
            return ({}, 403)
        else:
            return {}

    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": put_forbidden}}}}
    )
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status={"foo": "bar"},
    )
    assert not success
    assert not changed


def test_set_status_no_api_version():
    """Make sure that setting a status without api_version works as expected"""
    dm = setup_testable_manager(
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    )
    log.debug("Setting Status")
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version=None,
        status={"foo": "bar"},
    )
    assert success
    assert changed


#########################################
## set_status w/ manage_ansible_status ##
#########################################


def test_set_status_ansible_status_not_ready():
    """Make sure that when ansible status management is enabled and the status
    is not ready, the correct ansible status is injected.
    """
    dm = setup_testable_manager(
        manage_ansible_status=True,
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}},
    )
    status = oper8_status.make_application_status(
        ready_reason=oper8_status.ReadyReason.INITIALIZING
    )
    log.debug3("Input status: %s", status)
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status=status,
    )
    assert success
    assert changed  # Status updated
    success, content = dm.get_object_current_state(
        kind="Foo", name="bar", namespace="test"
    )
    assert success
    cluster_status = content["status"]
    log.debug3("Updated status: %s", cluster_status)
    ansible_conditions = get_ansible_conditions(cluster_status)
    assert len(ansible_conditions) == 1
    assert (
        ansible_conditions[0].get("ansibleResult")
        == OpenshiftDeployManager._ANSIBLE_COND_RES_UNREADY
    )


def test_set_status_ansible_status_ready():
    """Make sure that when ansible status management is enabled and the status
    is ready, the correct ansible status is injected.
    """
    dm = setup_testable_manager(
        manage_ansible_status=True,
        cluster_state={"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}},
    )
    status = oper8_status.make_application_status(
        ready_reason=oper8_status.ReadyReason.STABLE,
        updating_reason=oper8_status.UpdatingReason.STABLE,
        version="1.0.0",
    )
    log.debug3("Input status: %s", status)
    success, changed = dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status=status,
    )
    assert success
    assert changed  # Status updated
    success, content = dm.get_object_current_state(
        kind="Foo", name="bar", namespace="test"
    )
    assert success
    cluster_status = content["status"]
    log.debug3("Updated status: %s", cluster_status)
    ansible_conditions = get_ansible_conditions(cluster_status)
    assert len(ansible_conditions) == 1
    assert (
        ansible_conditions[0].get("ansibleResult")
        == OpenshiftDeployManager._ANSIBLE_COND_RES_READY
    )


def test_set_status_ansible_status_transition():
    """Make sure that when ansible status management is enabled and the status
    is ready when the previous status was not, the transition time is updated.
    """
    version = "1.0.0"
    init_status = oper8_status.make_application_status(
        ready_reason=oper8_status.ReadyReason.INITIALIZING, version=version
    )
    ready_status = oper8_status.make_application_status(
        ready_reason=oper8_status.ReadyReason.STABLE,
        updating_reason=oper8_status.UpdatingReason.STABLE,
        version=version,
    )
    dm = setup_testable_manager(
        manage_ansible_status=True,
        cluster_state={
            "test": {"Foo": {"foo.bar.com/v1": {"bar": {"status": init_status}}}}
        },
    )

    dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status=ready_status,
    )
    cluster_status = dm.get_object_current_state(
        kind="Foo", name="bar", namespace="test"
    )[1]["status"]
    log.debug3("Updated status: %s", cluster_status)
    ansible_conditions = get_ansible_conditions(cluster_status)
    assert len(ansible_conditions) == 1
    assert (
        ansible_conditions[0].get("ansibleResult")
        == OpenshiftDeployManager._ANSIBLE_COND_RES_READY
    )

    # Make sure timestamp taken from ready status
    assert (
        ansible_conditions[0].get("lastTransitionTime")
        == ready_status["conditions"][0][oper8_status.TIMESTAMP_KEY]
    )


def test_set_status_ansible_status_no_transition():
    """Make sure that when ansible status management is enabled and the status
    is ready when the previous status was also ready, the transition time is
    not updated.
    """
    init_status = oper8_status.make_application_status(
        ready_reason=oper8_status.ReadyReason.INITIALIZING
    )
    error_status = oper8_status.make_application_status(
        ready_reason=oper8_status.ReadyReason.CONFIG_ERROR
    )
    dm = setup_testable_manager(
        manage_ansible_status=True,
        cluster_state={
            "test": {"Foo": {"foo.bar.com/v1": {"bar": {"status": init_status}}}}
        },
    )

    dm.set_status(
        kind="Foo",
        name="bar",
        namespace="test",
        api_version="foo.bar.com/v1",
        status=error_status,
    )
    cluster_status = dm.get_object_current_state(
        kind="Foo", name="bar", namespace="test"
    )[1]["status"]
    log.debug3("Updated status: %s", cluster_status)
    ansible_conditions = get_ansible_conditions(cluster_status)
    assert len(ansible_conditions) == 1
    assert (
        ansible_conditions[0].get("ansibleResult")
        == OpenshiftDeployManager._ANSIBLE_COND_RES_UNREADY
    )

    # Make sure timestamp taken from first init status
    assert (
        ansible_conditions[0].get("lastTransitionTime")
        == init_status["conditions"][0][oper8_status.TIMESTAMP_KEY]
    )


###########
## Other ##
###########


def test_default_construct_client():
    """Make sure that if no client is given at construction, the client is
    automatically created when needed.
    """
    dm = OpenshiftDeployManager()
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    with mock_kub_client_constructor():
        dm.deploy(make_obj_states(cluster_state))


def test_in_cluster_config():
    """Make sure that if the in-cluster config loads, the client is created
    when needed."""
    dm = OpenshiftDeployManager()
    cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": {}}}}}
    with mock.patch("kubernetes.config.load_incluster_config", return_value=None):
        mock_client = MockKubClient()
        with mock.patch("kubernetes.client.ApiClient", return_value=mock_client):
            dm.deploy(make_obj_states(cluster_state))


def test_422_fallback_to_put_ok():
    """Make sure that if enabled, deploy can fall back to using PUT instead of
    PATCH on a 422 error and that the PUT can succeed if valid.
    """
    with library_config(deploy_unprocessable_put_fallback=True):

        def fail_patch_only(method, namespace, kind, api_version, name):
            if method == "PATCH":
                return ({}, 422)
            return {
                "apiVersion": api_version,
                "kind": kind,
                "metadata": {
                    "name": name,
                    "namespace": namespace,
                },
            }

        cluster_state = {"test": {"Foo": {"foo.bar.com/v1": {"bar": fail_patch_only}}}}
        dm = setup_testable_manager(cluster_state=cluster_state)
        success, _ = dm.deploy(
            [
                {
                    "apiVersion": "foo.bar.com/v1",
                    "kind": "Foo",
                    "metadata": {"name": "bar", "namespace": "test"},
                    "key": "val",
                }
            ]
        )
        assert success


def test_422_fallback_to_put_invalid():
    """Make sure that if enabled, deploy can fall back to using PUT instead of
    PATCH on a 422 error and that the PUT can still produce a 422 if it is truly
    invalid.
    """
    with library_config(deploy_unprocessable_put_fallback=True):

        def fail_put_patch_only(method, namespace, kind, api_version, name):
            if method in ["PATCH", "PUT"]:
                return ({}, 422)
            return {
                "apiVersion": api_version,
                "kind": kind,
                "metadata": {
                    "name": name,
                    "namespace": namespace,
                },
            }

        cluster_state = {
            "test": {"Foo": {"foo.bar.com/v1": {"bar": fail_put_patch_only}}}
        }
        dm = setup_testable_manager(cluster_state=cluster_state)
        success, _ = dm.deploy(
            [
                {
                    "apiVersion": "foo.bar.com/v1",
                    "kind": "Foo",
                    "metadata": {"name": "bar", "namespace": "test"},
                    "key": "val",
                }
            ]
        )
        assert not success



#!/usr/bin/python
# -*- coding: utf-8 -*-

# NOTE: For ansible, the shebang must be an absolute path, not env
# https://docs.ansible.com/ansible/latest/dev_guide/developing_modules_documenting.html#python-shebang-utf-8-coding

## Imports #####################################################################

# Standard
import ast
import logging.handlers
import pathlib

# COMPATIBILITY
# With >= 3.12, the vendored ansible six functionality does not import cleanly,
# so we patch sys.modules to work around this
import sys

if sys.version_info.major > 3 or sys.version_info.minor > 11:
    # Third Party
    from six import moves

    sys.modules["ansible.module_utils.six.moves"] = moves

# Third Party
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.k8s.common import K8sAnsibleMixin

# First Party
import alog

# Local
from oper8 import config
from oper8.reconcile import ReconcileManager
from oper8.watch_manager.ansible_watch_manager.modules.log_rotator import (
    AutoRotatingFileHandler,
)

## Module Doc Vars #############################################################

ANSIBLE_METADATA = {
    "metadata_version": "0.1",
    "status": ["preview"],
    "supported_by": "oper8.org",
}

DOCUMENTATION = """
---
module: k8s_application

short_description: Ansible module which manages a Controller using the
    oper8 python library

version_added: "2.4"

description:
    - "TODO"

options:

    version:
        description:
            - This dictates the version of the application library which will
              be used to service the CR
        type: str
        required: true
    controller_class:
        description:
            - This is the fully-qualified name of the python Controller class
              that is bound to this operator.
        type: str
        required: true
    operation:
        description:
            - Which operation this execution should perform (add or remove)
        type: str ("add" or "remove")
        required: false
    full_cr:
        description:
            - This template passes the content of the full CR which triggered
              the deploy into the library so that the deploy parameters can be
              parsed there. It should be populated with a template expression
              based on the group and kind of the CR being managed.
        type: str (template)
        required: true
    manage_ansible_status:
        description:
            - If true, oper8 will emulate the status management done natively by
              ansible based on the readiness values of oper8's native status
              management
        type: str
        required: false
    enable_ansible_vcs:
        description: Whether to enable or disable ansible-vcs
        type: bool
        default: False
    vcs_dir:
        description: The directory holding the git repo used with ansible-vcs
        type: str
        required: false
    vcs_dest:
        description: The destination directory for the checkedout version used with ansible-vcs
        type: str
        required: false
    strict_versioning:
        description:
            - If true, oper8 will reject spec.version values that are not in the
              list of supported_versions.
        required: false

    log_level:
        description:
            - This sets the default verbosity level of the logging from inside
              of the oper8 library.
        type: str
        required: false
    log_filters:
        description:
            - This sets per-channel verbosity levels for the logging from inside
              of the oper8 library.
        type: str
        required: false
    log_json:
        description:
            - This sets the output formatter for oper8 logging to be
              'json' rather than 'pretty'
        type: bool
        required: false
    log_file:
        description:
            - This adds a file path to send log output to
        type: str
        required: false
    log_thread_id:
        description:
            - This adds the unique thread id to each log message
        type: bool
        required: false

    dry_run:
        description:
            - This sets the operator to render-only mode and will not actually
              manage the downstream resources.
        type: bool
        required: false
    standalone:
        description:
            - This sets the oper8 library to operate outside of
              'operator' mode, enabling the playbook to be run directly from the
              command line rather than inside of a deployed operator.
        type: bool
        required: false
    working_dir:
        description:
            - This sets a named working directory to be used by the
              oper8 library rather than letting it create a random
              ephemeral working dir internally.
            - WARNING: Setting a working_dir is for debugging ONLY! It will
              leave the generated resource files behind which can cause
              problems if the set of files and/or names changes between deploys.
        type: str
        required: false

author:
    - Gabe Goodhart (@gabe.l.hart)
"""

EXAMPLES = """
# Deploy the mlflow application
- name: MLFlow Controller
  k8s_application:
    controller_class: mlflow_application.application.mlflow.MLFlowController
    full_cr: "{{ _org_oper8_mlflow_mlflow }}"

"""

RETURN = """
success:
    description: Whether or not the deployment rolled out successfully
    type: bool
    returned: always
log_output:
    description: Full string of log messages produced during the run
    type: str
    returned: always
should_requeue:
    description: Whether or not the reconcile request should be re-queued
    type: bool
    returned: always
requeue_after:
    description: Tells the controller runtime to requeue the reconcile request after the duration
    type: string
    returned: always
    sample: '60s'
"""


## Module Class ################################################################

log = alog.use_channel("K8S-APP")


class KubernetesAnsibleApplicationModule(AnsibleModule, K8sAnsibleMixin):
    """This ansible module implements management of a full application via
    the oper8 modules.
    """

    @property
    def argspec(self):
        """Inherit the argspec for a K8sAnsibleMixin and add any necessary
        overrides
        """
        spec = {}

        # Add an argument for the version. This is needed so that the operator
        # can pass its version through to sub-CRs
        spec["version"] = dict(type="str", required=True)

        # Add an argument to specify the python class to use for the application
        spec["controller_class"] = dict(type="str", required=True)

        # Determine whether this is running as a finalizer (remove) or
        # reconciler (add)
        spec["operation"] = dict(type="str", default="add")

        # If true, oper8 will emulate the status management done natively by
        # ansible based on the readiness values of oper8's native status
        # management
        spec["manage_ansible_status"] = dict(type="bool", default=False)

        # The directory where ansible-vcs sources its version checkouts
        spec["vcs_dir"] = dict(type="str", default="/opt/ansible/app/")
        spec["vcs_dest"] = dict(
            type="str", default="/opt/ansible/app/version/{version}/{kind}/{name}/"
        )
        spec["enable_ansible_vcs"] = dict(type="bool", default=False)

        # If true, oper8 will require that a list of supported_versions is found
        # and that the spec.version value is in the list
        spec["strict_versioning"] = dict(type="bool", default=False)

        # Add arguments to dry run the deploy and specify the working dir
        spec["dry_run"] = dict(type="bool", default=False)
        spec["standalone"] = dict(type="bool", default=False)
        spec["working_dir"] = dict(type="str", default=config.working_dir or None)

        # Add arguments for alog configuration
        spec["log_level"] = dict(type="str", default="")
        spec["log_filters"] = dict(type="str", default="")
        spec["log_json"] = dict(type="bool", default=False)
        spec["log_thread_id"] = dict(type="bool", default=False)
        spec["log_file"] = dict(type="str", default=None)

        # This parameter is where the full CR spec will live
        spec["full_cr"] = dict(type="str", required=True)

        # Return the fully constructed spec
        return spec

    def __init__(self, *args, **kwargs):
        """Passthrough constructor"""
        kwargs["argument_spec"] = self.argspec
        super().__init__(*args, **kwargs)

    def execute_module(self):
        """Execute the constructed module by parsing the resource config"""

        # Parse the full CR content
        try:
            full_cr = ast.literal_eval(self.params["full_cr"])
        except (ValueError, SyntaxError) as e:
            return self.fail_json(
                msg="Failed to parse full_cr", err=str(e), success=False
            )

        # BACKWARDS COMPATIBILITY -- For most parameters the configurations are
        # now library-level via the config package. To avoid breaking existing
        # workflows, we allow them to be overridden for each reconciliation. In
        # reality, they will always be set the same since they are generally
        # static for a given invocation of the operator.
        config.dry_run = self.params["dry_run"]
        config.standalone = self.params["standalone"]
        config.working_dir = self.params.get("working_dir", config.working_dir)
        config.strict_versioning = self.params["strict_versioning"]

        # Configure VCS config
        if self.params["enable_ansible_vcs"]:
            config.vcs.enabled = True
            if not config.vcs.version_override and self.params["version"]:
                config.vcs.version_override = self.params["version"]

        # If VCS dir or dest weren't overridden by config than supply the
        # default ansible directory and destination
        if not config.vcs.repo:
            config.vcs.repo = self.params["vcs_dir"]
        if not config.vcs.dest:
            config.vcs.dest = self.params["vcs_dest"]

        # Configure Logging values
        if self.params["log_level"]:
            config.log_level = self.params["log_level"]
        if self.params["log_filters"]:
            config.log_filters = self.params["log_filters"]
        if self.params["log_json"]:
            config.log_json = self.params["log_json"]
        if self.params["log_thread_id"]:
            config.log_thread_id = self.params["log_thread_id"]

        # Add the Ansible log formatter
        if self.params["log_file"] is not None:
            # Ensure log directory exists
            log_file_path = pathlib.Path(self.params["log_file"])
            log_file_path.parent.mkdir(parents=True, exist_ok=True)

            # Remove all of the current handlers as STDOUT is captured by ansible
            # thus can't be used for logging
            root_logger = logging.getLogger()
            root_logger.handlers.clear()

            handler = AutoRotatingFileHandler(self.params["log_file"])
            handler.setFormatter(alog.AlogPrettyFormatter())
            root_logger.addHandler(handler)

        # Invoke the reconcile
        is_finalizer = self.params["operation"] == "remove"
        reconcile_manager = ReconcileManager()
        reconcile_result = reconcile_manager.safe_reconcile(
            self.params["controller_class"], full_cr, is_finalizer
        )

        # Return the success value
        requeue_seconds = reconcile_result.requeue_params.requeue_after.total_seconds()
        return self.exit_json(
            success=True,
            should_requeue=reconcile_result.requeue,
            requeue_after=f"{round(requeue_seconds)}s",
        )


## Main ########################################################################


def main():
    KubernetesAnsibleApplicationModule().execute_module()


if __name__ == "__main__":  # pragma: no cover
    main()



"""
Module to validate values in a loaded config
"""

# Standard
from typing import Any, Dict, List, Optional, Union
import abc
import builtins

# First Party
import aconfig
import alog

# Local
from .. import constants
from ..utils import nested_get  # pylint: disable=cyclic-import

log = alog.use_channel("CONFG")


################################################################################
## Public ######################################################################
################################################################################


def get_invalid_params(
    config: aconfig.Config,
    validation_config: aconfig.Config,
) -> List[str]:
    """Get a list of any params that are invalid

    Args:
        config:  aconfig.Config
            The parsed config with any override values
        validation_config:  aconfig.Config
            The parallel config holding validation setup

    Returns:
        invalid_params:  List[str]
            A list of all string keys for parameters that fail validation
    """

    # For each validation element, perform the validation
    invalid_params = []
    for val_key, validator in _parse_validation_config(validation_config).items():
        if not validator.validate(nested_get(config, val_key)):
            log.warning("Found invalid config key [%s]", val_key)
            invalid_params.append(val_key)

    # Return the list of invalid params
    return invalid_params


################################################################################
## Implementation ##############################################################
################################################################################


## Base Class ##################################################################

# NOTE: Pylint dislikes classes with a single public member function, but
#   they're useful here to define the inheritance structure, so we disable this
#   warning and re-enable it later.
#
# pylint: disable=too-few-public-methods


class _ValidatedParameter(abc.ABC):
    """This class represents a parameter with type and value validation"""

    def __init__(
        self,
        valid_types: List[type],
        optional: bool = False,
    ):
        """Construct with the set of valid types

        Args:
            valid_types:  List[type]
                The list of valid types for this parameter
            optional:  bool
                Whether or not the parameter is optional
        """
        assert len(valid_types) > 0, "Must specify at least one valid type"
        assert all(
            isinstance(typ, type) for typ in valid_types
        ), "Got a non-type in valid_types"
        self.valid_types = valid_types
        self.optional = optional

    def validate(self, value: Any) -> bool:
        """Run the validation for a read value

        Args:
            value:  Any
                The value to validate against this parameter

        Returns:
            valid:  bool
                True if the value is valid, False otherwise
        """
        # If the parameter is optional and the value is None, it's valid
        if self.optional and value is None:
            return True

        # Make sure the type is valid
        if not any(isinstance(value, valid_type) for valid_type in self.valid_types):
            log.warning("Invalid type <%s>", type(value))
            return False

        # Make sure the value is valid
        valid_value = self._validate_value(value)
        if not valid_value:
            log.warning("Invalid value [%s]", value)
        return valid_value

    @abc.abstractmethod
    def _validate_value(self, value: Any) -> bool:
        """All child classes must provide value validation that is specific to
        the given type
        """


## Validators ##################################################################


############
## number ##
############


class _NumberParameter(_ValidatedParameter):
    """A parameter that must be a number type and has optional bounds"""

    TYPES = [int, float]
    TYPE_KEY = "number"

    def __init__(
        self,
        *,
        min: Optional[Union[int, float]] = None,  # pylint: disable=redefined-builtin
        max: Optional[Union[int, float]] = None,  # pylint: disable=redefined-builtin
        **kwargs,
    ):
        """Construct with optional bounds

        NOTE: The builtin min/max names are used here so that the arguments have
            the appropriate intuitive names in the configuration yaml file

        Kwargs:
            min:  Optional[Union[int, float]]
                If not None, minimum value (inclusive) allowed
            max:  Optional[Union[int, float]]
                If not None, maximum value (inclusive) allowed
        """
        super().__init__(valid_types=self.TYPES, **kwargs)
        self._min = min
        self._max = max

    def _validate_value(self, value: Union[int, float]) -> bool:
        """Validate the value against the configured bounds"""
        return (self._min is None or value >= self._min) and (
            self._max is None or value <= self._max
        )


#########
## int ##
#########


class _IntParameter(_NumberParameter):
    """A number parameter that must be an int"""

    TYPES = [int]
    TYPE_KEY = "int"

    def __init__(self, *_, **kwargs):
        super().__init__(**kwargs)


###########
## float ##
###########


class _FloatParameter(_NumberParameter):
    """A number parameter that must be an float"""

    TYPES = [float]
    TYPE_KEY = "float"

    def __init__(self, *_, **kwargs):
        super().__init__(**kwargs)


#########
## str ##
#########


class _StrParameter(_ValidatedParameter):
    """A parameter that must be of type str and has optional length bounds"""

    TYPES = [str]
    TYPE_KEY = "str"

    def __init__(
        self,
        *,
        min_len: Optional[int] = None,
        max_len: Optional[int] = None,
        **kwargs,
    ):
        """Construct with optional length bounds

        Kwargs:
            min_len:  Optional[int]
                If not None, minimum length (inclusive) allowed
            max_len:  Optional[int]
                If not None, maximum length (inclusive) allowed
        """
        super().__init__(valid_types=self.TYPES, **kwargs)
        self._min_len = min_len
        self._max_len = max_len

    def _validate_value(self, value: str) -> bool:
        """Validate the value against the configured length bounds"""
        return (self._min_len is None or len(value) >= self._min_len) and (
            self._max_len is None or len(value) <= self._max_len
        )


##########
## bool ##
##########


class _BoolParameter(_ValidatedParameter):
    """A parameter that must be a bool"""

    TYPES = [bool]
    TYPE_KEY = "bool"

    def __init__(self, *_, **kwargs):
        """Pass through constructor"""
        super().__init__(valid_types=self.TYPES, **kwargs)

    def _validate_value(self, value: bool) -> bool:
        """Value validation is a no-op for bool"""
        return True


##########
## enum ##
##########


class _EnumParameter(_ValidatedParameter):
    """A parameter with a fixed set of valid str or int values"""

    TYPES = [str, int, type(None)]
    TYPE_KEY = "enum"

    def __init__(self, *, values: List[Union[str, int, type(None)]], **kwargs):
        """Construct with the list of enum values

        Args:
            values:  List[Union[str, int, type(None)]]
                The list of valid values for this enum. Note that it can be a
                mix of any of the valid value types
        """
        super().__init__(valid_types=self.TYPES, **kwargs)
        assert (
            isinstance(values, list) and values
        ), "Must specify at least one enum value!"
        self.values = values

    def _validate_value(self, value: Union[str, int, type(None)]) -> bool:
        """Make sure the given value is in the configured value set"""
        return value in self.values


##########
## list ##
##########


class _ListParameter(_ValidatedParameter):
    """A parameter that must be of type list and has optional constraints around
    the type and count of elements
    """

    TYPES = [list]
    TYPE_KEY = "list"

    def __init__(
        self,
        *,
        min_len: Optional[int] = None,
        max_len: Optional[int] = None,
        item_type: Optional[str] = None,
        **kwargs,
    ):
        """Construct with optional length bounds

        Kwargs:
            min_len:  Optional[int]
                Minimum number of items that must be provided
            max_len:  Optional[int]
                Maximum number of items that must be provided
            item_type:  Optional[str]
                String name of the type that the items must be (e.g. "str")
        """
        super().__init__(valid_types=self.TYPES, **kwargs)
        self._min_len = min_len
        self._max_len = max_len
        self._item_type = None
        if item_type is not None:
            assert hasattr(builtins, item_type), f"Unsupported item_type: {item_type}"
            self._item_type = getattr(builtins, item_type)
            assert (
                self._item_type is not None
            ), f"Failed to fetch item_type: {item_type}"

    def _validate_value(self, value: str) -> bool:
        """Validate the value against the configured length bounds and item type"""
        return (
            (self._min_len is None or len(value) >= self._min_len)
            and (self._max_len is None or len(value) <= self._max_len)
            and (
                self._item_type is None
                or all(isinstance(item, self._item_type) for item in value)
            )
        )


# Re-enable the pylint warning
# pylint: enable=too-few-public-methods

## Factory #####################################################################


def _create_factory_map(param_class, factory_map=None):
    """Helper to recursively create the singleton factory map"""
    factory_map = factory_map or {}

    # Add this class if it's not abstract
    if not param_class.__abstractmethods__:
        factory_map[param_class.TYPE_KEY] = param_class

    # Recurse
    for subclass in param_class.__subclasses__():
        factory_map = _create_factory_map(subclass, factory_map)

    return factory_map


# Global map from type keys to parameter type classes
_factory_map = _create_factory_map(_ValidatedParameter)


def _construct_parameter(param_args: Dict[str, Any]) -> _ValidatedParameter:
    """Construct a _ValidatedParameter from the given args parsed out of a
    validation file.

    Args:
        param_args:  Dict[str, Any]
            The key/value pairs for this parameter

    Returns:
        parameter:  Optional[_ValidatedParameter]
            The constructed _ValidatedParameter based on the given args if the
            type is known. If unkonwn, None is returned.
    """
    assert "type" in param_args, "All parameters must have a 'type'"
    param_type = param_args["type"]
    if not (isinstance(param_type, str) and param_type in _factory_map):
        return None
    param_args.pop("type")
    return _factory_map[param_type](**param_args)


## Parsing #####################################################################


def _parse_validation_config(
    validation_config: aconfig.Config,
    prefix_parts: Optional[List[str]] = None,
) -> Dict[str, _ValidatedParameter]:
    """Recursively parse the given validation file into a dict of nested keys
    pointing to _ValidatedParameter instances.
    """
    output_dict = {}
    prefix_parts = prefix_parts or []
    for key, val in validation_config.items():
        assert isinstance(key, str), "Only string keys allowed!"
        if isinstance(val, dict):
            # Make the nested key
            key_parts = prefix_parts + [key]
            nested_key = constants.NESTED_DICT_DELIM.join(key_parts)

            # If the dict has a "type" field, try parsing it as a validated
            # parameter
            param = None
            if "type" in val:
                log.debug3(
                    "Attempting to construct parameter at [%s]: %s", nested_key, val
                )
                param = _construct_parameter(val)

            # If a valid param was parsed, update the config
            if param:
                log.debug3("Found parameter at %s", nested_key)
                output_dict[nested_key] = param

            # Otherwise, recurse
            else:
                log.debug3("Recursing into %s", nested_key)
                output_dict.update(
                    _parse_validation_config(
                        val,
                        prefix_parts=key_parts,
                    )
                )

    return output_dict



"""
This module holds all of the command classes for oper8's main entrypoint
"""

# Local
from .base import CmdBase
from .check_heartbeat import CheckHeartbeatCmd
from .run_operator_cmd import RunOperatorCmd
from .setup_vcs_cmd import SetupVCSCmd



"""
Tests for the Leadership common functions
"""
# Third Party
import pytest

# Local
from oper8.test_helpers.helpers import library_config
from oper8.watch_manager.python_watch_manager.leader_election import (
    AnnotationLeadershipManager,
    DryRunLeadershipManager,
    LeaderForLifeManager,
    LeaderWithLeaseManager,
    get_leader_election_class,
)

## Helpers #####################################################################


@pytest.mark.parametrize(
    ["config", "expected_class"],
    [
        ["leader-for-life", LeaderForLifeManager],
        ["leader-with-lease", LeaderWithLeaseManager],
        ["annotation", AnnotationLeadershipManager],
        ["dryrun", DryRunLeadershipManager],
    ],
)
def test_get_leader_election_class(config, expected_class):
    with library_config(python_watch_manager={"lock": {"type": config}}):
        assert get_leader_election_class() == expected_class



"""
Test the DAG Graph Class functionality
"""
# Third Party
import pytest

# Local
from oper8.dag import Graph, Node

################################################################################
## Graph Tests #############################################################
################################################################################


def test_creation():
    """Test node initialization"""
    graph = Graph()
    assert graph.root


def test_add_node():
    """Test node children functions"""
    graph = Graph()
    node_a = Node("a")
    graph.add_node(node_a)
    assert graph.root.has_child(node_a)


def test_add_node_exception():
    """Test node children functions"""
    graph = Graph()
    with pytest.raises(ValueError):
        graph.add_node(Node(None))

    graph.add_node(Node("a"))
    with pytest.raises(ValueError):
        graph.add_node(Node("a"))


def test_add_node_dependency():
    """Test node children functions"""
    graph = Graph()
    node_a = Node("a")
    node_b = Node("b")
    node_c = Node("c")
    node_d = Node("d")
    graph.add_node(node_a)
    graph.add_node(node_b)
    graph.add_node(node_c)
    graph.add_node(node_d)
    graph.add_node_dependency(node_a, node_b)
    graph.add_node_dependency(node_a, node_c, "testdata")
    graph.add_node_dependency(node_b, node_c)
    graph.add_node_dependency(node_c, node_d)

    assert node_a.get_children() == [(node_b, None), (node_c, "testdata")]


def test_add_node_dependency_exception():
    """Test node children functions"""
    graph = Graph()
    node_a = Node("a")
    node_b = Node("b")
    graph.add_node(node_a)
    graph.add_node(node_b)

    with pytest.raises(ValueError):
        graph.add_node_dependency(node_a, Node("c"))
    with pytest.raises(ValueError):
        graph.add_node_dependency(Node("c"), node_a)


def test_accessors():
    """Test node children functions"""
    graph = Graph()
    node_a = Node("a")
    node_b = Node("b")
    graph.add_node(node_a)
    graph.add_node(node_b)

    assert graph.get_node("a") == node_a
    assert graph.get_all_nodes() == [node_a, node_b]
    assert graph.has_node(node_a)
    assert node_a in graph
    assert not graph.has_node(Node("c"))


def test_topology():
    graph = Graph()
    node_a = Node("a")
    node_b = Node("b")
    graph.add_node(node_a)
    graph.add_node(node_b)
    graph.add_node_dependency(node_b, node_a)

    assert graph.topology() == [node_a, node_b]


def test_topology_ordering():
    """Test node topology function"""
    graph = Graph()
    node_a = Node("a")
    node_b = Node("b")
    node_c = Node("c")
    graph.add_node(node_c)
    graph.add_node(node_b)
    graph.add_node(node_a)

    assert graph.topology() == [node_a, node_b, node_c]


def test_str():
    graph = Graph()
    node_a = Node("a")
    node_b = Node("b")
    graph.add_node(node_a)
    graph.add_node(node_b)
    graph.add_node_dependency(node_a, node_b)

    assert str(graph) == "Graph({a:[b],b:[]})"



"""
Version Control System class manages a specific git directory.
"""

# Standard
from enum import Enum
from typing import List, Optional, Set, Tuple
import pathlib
import time

# Third Party
from pygit2 import (  # pylint: disable=no-name-in-module
    GIT_CONFIG_LEVEL_GLOBAL,
    GIT_OPT_GET_SEARCH_PATH,
    AlreadyExistsError,
    Branch,
    Commit,
    Config,
    GitError,
    Reference,
    Repository,
    Signature,
    clone_repository,
    init_repository,
    option,
)

# First Party
import alog

# Local
from .exceptions import ConfigError, Oper8FatalError, PreconditionError

log = alog.use_channel("VCS")


## VCS Constants #############################################################


class VCSCheckoutMethod(Enum):
    """Enum for available VCS checkout methods"""

    WORKTREE = "worktree"
    CLONE = "clone"


## VCS Errors #############################################################


class VCSMultiProcessError(PreconditionError):
    """VCS Error for when multiple git processes attempt to update the git directory
    at the same time"""


class VCSConfigError(ConfigError):
    """Error for VCS Specific config exception"""


class VCSRuntimeError(Oper8FatalError):
    """Error for general git exceptions"""


## VCS  #############################################################


class VCS:
    """Generic class for handling a git repository. This class contains helper functions
    to get, list, and checkout references. Each instance of this class corresponds to a
    different git directory
    """

    def __init__(self, directory: str, create_if_needed: bool = False, **kwargs):
        """Initialize the pygit2 Repository reference

        Args:
            directory: str
                The git directory
            create_if_needed: bool
                If True, the repo will be initialized if it doesn't already
                exist
            **kwargs:
                Passthrough args to the repository setup
        """
        # Get repo reference
        try:
            # Check for global file and create one if needed. This
            # is needed due to this issue: https://github.com/libgit2/pygit2/issues/915
            config_file = (
                pathlib.Path(option(GIT_OPT_GET_SEARCH_PATH, GIT_CONFIG_LEVEL_GLOBAL))
                / ".gitconfig"
            )
            if not config_file.exists():
                config_file.touch(exist_ok=True)

            # Disable safe git directories. This solves a common problem
            # when running in openshift where the running user is different
            # from the owner of the filesystem
            global_config = Config.get_global_config()
            global_config["safe.directory"] = "*"

            self.repo = Repository(directory)
            log.debug2("Found repo: %s", self.repo)
        except GitError as err:
            if create_if_needed:
                self.repo = init_repository(directory, **kwargs)
            else:
                log.error("Invalid Repo: %s", err, exc_info=True)
                raise VCSConfigError(f"Invalid Repo at {directory}") from err

    ### Accessors

    @property
    def head(self) -> str:
        """Get a reference to the current HEAD"""
        return str(self.repo.head.target)

    def get_ref(self, refish: str) -> Tuple[Commit, Reference]:
        """Get a git commit and reference from a shorthand string

        Args:
            refish: str
                The human readable form of a git reference like branch name
                or commit hash

        Returns
            commit_and_reference: Tuple[Commit,Reference]
                Both a commit and reference for a given refish
        """
        try:
            return self.repo.resolve_refish(refish)
        except KeyError as err:
            log.error("Unable to find version %s in repo", refish)
            raise VCSConfigError(  # pylint: disable=raise-missing-from
                f"Version: '{refish}' not found in repo"
            ) from err

    def list_refs(self) -> Set[str]:
        """List all of the tags and references in the repo

        Returns
            ref_list: Set[str]
                A set of all references' shorthand as strings
        """
        # Loop through repo tags to get each tag's short name
        refs_set = set()
        for ref in self.repo.references.objects:
            refs_set.add(ref.shorthand)

        return refs_set

    ### Mutators

    def checkout_ref(
        self,
        refish: str,
        dest_path: Optional[pathlib.Path] = None,
        method: VCSCheckoutMethod = VCSCheckoutMethod.WORKTREE,
        **kwargs,
    ):
        """Checkout a refish to a given destination directory. This function
        first attempts to create a worktree but on failure will do a traditional
        clone

        Args:
            refish: str
                The refish to be checked out in the dest_dir
            dest_path: Optional[pathlib.Path]
                The destination directory if not in-place
            method: VCSCheckoutMethod=VCSCheckoutMethod.WORKTREE
                The checkout method to use, either a git clone or worktree add
            **kwargs
                Kwargs to pass through to checkout
        """

        # Get the commit and ref for a given refish
        commit, ref = self.get_ref(refish)

        # If in-place, check out directly
        if not dest_path:
            log.debug2("Checking out %s in place", refish)
            self.repo.checkout(ref, **kwargs)
            return

        # Check if dest directory already exists and if it has the correct
        # commit
        if dest_path.is_dir():
            dest_vcs = VCS(dest_path)

            # Check if the dest index file has been created. It is the last
            # part of a checkout. If index has not been created than another
            # process must be working on it
            dest_index_file = pathlib.Path(dest_vcs.repo.path) / "index"
            if not dest_index_file.is_file():
                raise VCSMultiProcessError(
                    "Index file not found. Checkout already in progress "
                )

            if dest_vcs.repo.head.peel(Commit) != commit:
                raise VCSConfigError(
                    f"Destination directory {dest_path} already exists with incorrect branch"
                )
            return

        # Create the directory if it doesn't exist
        dest_path.parents[0].mkdir(parents=True, exist_ok=True)

        if method == VCSCheckoutMethod.WORKTREE:
            # Create a unique branch for each worktree
            cleaned_dest_dir = "_".join(dest_path.parts[1:])
            branch_name = f"{refish}_{cleaned_dest_dir}"

            branch = self.create_branch(branch_name, commit)
            self._create_worktree(branch_name, dest_path, branch)
        elif method == VCSCheckoutMethod.CLONE:
            self._clone_ref(dest_path, ref, **kwargs)
        else:
            raise VCSConfigError(f"Invalid checkout method: {method}")

    def create_commit(
        self,
        message: str,
        parents: Optional[List[str]] = None,
        committer_name: str = "Oper8",
        committer_email: str = "noreply@oper8.org",
    ):
        """Create a commit in the repo with the files currently in the index

        Args:
            message: str
                The commit message
            parents: Optional[List[str]]
                Parent commit hashes
            committer_name: str
                The name of the committer
            committer_email: str
                Email address for this committer
        """
        parents = parents or []
        parent_commits = []
        for parent in parents:
            try:
                parent_commits.append(self.repo.get(parent))
            except ValueError as err:
                raise ValueError(f"Invalid parent commit: {parent}") from err
        signature = Signature(committer_name, committer_email)
        self.repo.create_commit(
            "HEAD", signature, signature, message, self.repo.index.write_tree(), parents
        )

    def add_remote(self, remote_name: str, remote_path: str):
        """Add a named remote to the repo

        Args:
            remote_name: str
                The name of the remote
            remote_path: str
                The path on disk to the remote repo
        """
        self.repo.remotes.create(remote_name, remote_path)

    def delete_remote(self, remote_name: str):
        """Remove a remote from the repo

        Args:
            remote_name:  str
                The name of the remote
        """
        self.repo.remotes.delete(remote_name)

    def fetch_remote(
        self,
        remote_name: str,
        refs: Optional[Set[str]] = None,
        wait: bool = True,
    ):
        """Fetch content from the named remote. If no refs given, all refs are
        fetched.

        Args:
            remote_name: str
                The name of the remote to fetch
            refs: Optional[Set[str]]
                The refs to fetch (fetch all if not given)
            wait: bool
                If true, wait for fetch to complete
        """
        remote = self.repo.remotes[remote_name]
        progress = remote.fetch(list(refs or []))
        while wait and progress.received_objects < progress.total_objects:
            time.sleep(0.1)  # pragma: no cover

    def create_branch(self, branch_name: str, commit: Commit) -> Branch:
        """Create branch given a name and commit

        Args:
            branch_name: str
                The name to be created
            commit: Commit
                The commit for the branch to be created from

        Returns:
            branch: Branch
                The created branch"""
        if branch_name in self.repo.branches:
            branch = self.repo.branches.get(branch_name)
            if branch.peel(Commit) != commit:
                raise VCSRuntimeError("Branch already exists with incorrect commit")
            return branch

        try:
            log.debug("Creating branch for %s", branch_name)
            return self.repo.branches.create(branch_name, commit)
        except AlreadyExistsError as err:
            # Branch must have been created by different processes
            log.warning("Branch %s already exists", branch_name)
            raise VCSMultiProcessError(f"Branch {branch_name} already exists") from err

        except OSError as err:
            raise VCSRuntimeError("Unable to create branch") from err

    def delete_branch(self, branch_name: str):
        """Delete a branch from the repo

        Args:
            branch_name:  str
                The name of the branch
        """
        self.repo.branches.delete(branch_name)

    def delete_tag(self, tag_name: str):
        """Delete a tag from the repo

        Args:
            tag_name:  str
                The name of the tag
        """
        self.repo.references.delete(f"refs/tags/{tag_name}")

    def checkout_detached_head(self, refish: Optional[str] = None):
        """Check out the current HEAD commit as a detached head

        Args:
            refish:  Optional[str]
                The ref to check out. If not given, the current HEAD is used
        """
        refish = refish or self.head

        # Create a placeholder reference to a non-existent remote
        dummy_ref = self.repo.references.create(
            "refs/remotes/doesnotexist/foobar", refish
        )
        self.repo.checkout(dummy_ref)
        self.repo.references.delete(dummy_ref.name)

    def compress_references(self):
        """Compress unreachable references in the repo"""
        self.repo.compress_references()

    ### Implementation Details

    def _clone_ref(self, dest_path: pathlib.Path, ref: Reference, **kwargs):
        """Clone a refish to a given destination directory

        Args:
            dest_path: pathlib.Path
                The destination directory
            refish: str
                The branch or ref to be checked out
            **kwargs
                Kwargs to pass through to checkout
        """
        try:
            dest_repo = clone_repository(self.repo.path, dest_path)
            dest_repo.checkout(refname=ref, **kwargs)
        except (OSError, GitError, KeyError) as err:
            log.error("Unable to clone refish: %s", ref.shorthand, exc_info=True)
            raise VCSRuntimeError("Unable to clone ref from repo") from err

    def _create_worktree(
        self, worktree_name: str, dest_path: pathlib.Path, branch: Branch
    ):
        """Create worktree for branch. This is better than a direct checkout
        as it saves space on checkout and is faster. This is especially
        beneficial on repositories with large git directories

        Args:
           worktree_name: str
               The name of the worktree
           dest_path: pathlib.Path
               The destination directory
           branch: Branch
               The branch to be checked out in the worktree
        """
        log.debug("Creating new worktree for %s", worktree_name)
        try:
            self.repo.add_worktree(worktree_name, dest_path, branch)
        except AlreadyExistsError as err:
            # Worktree must have been created by different processes
            log.warning("Worktree %s already exists", worktree_name)
            raise VCSMultiProcessError(
                f"Worktree {worktree_name} already exists"
            ) from err
        except GitError as err:
            # If reference is already checked out it must have been done by a different process
            if "is already checked out" in str(err):
                log.warning(
                    "Branch %s already checked out by other process",
                    worktree_name,
                    exc_info=True,
                )
                raise VCSMultiProcessError(
                    f"Branch {worktree_name} already checked out by other process"
                ) from err

            log.error(
                "Unexpected Git Error when adding worktree: %s", err, exc_info=True
            )
            raise VCSRuntimeError(
                "Adding worktree failed with unexpected git error"
            ) from err



"""
This module implements a factory for TlsContext implementations
"""

# Standard
from typing import Optional, Type

# First Party
import alog

# Local
from .interface import ITlsContext
from oper8 import Session, assert_config
from oper8.utils import merge_configs

log = alog.use_channel("TLSFY")


## Interface ###################################################################


def get_tls_context(
    session: Session,
    config_overrides: Optional[dict] = None,
) -> ITlsContext:
    """Get an instance of the configured implementation of the tls context

    Args:
        session:  Session
            The current deploy session
        config_overrides:  Optional[dict]
            Optional runtime config values. These will overwrite any values
            pulled from the session.config

    Returns:
        tls_context:  ITlsContext
            The constructed instance of the context
    """
    return _TlsContextSingletonFactory.get_tls_context(
        session,
        config_overrides=config_overrides,
    )


def register_tls_context_type(context_class: Type[ITlsContext]):
    """Register a constructor for a given context implementation type

    Args:
        context_class:  Type[ITlsContext]
            The ITlsContext child class to register
    """
    _TlsContextSingletonFactory.register(context_class)


## Implementation Details ######################################################


class _TlsContextSingletonFactory:
    """The _TlsContextSingletonFactory will manage a singleton instance of an
    ITlsContext based on the session's config.
    """

    # The section of the app_config that will hold the config
    _APP_CONFIG_SECTION = "tls"
    _CONFIG_TYPE_FIELD = "type"

    # Internal class dict holding the registered types
    _registered_types = {}

    # Singleton instance details
    _instance = None
    _instance_deploy_id = None

    ## Interface ###############################################################

    @classmethod
    def get_tls_context(
        cls,
        session: Session,
        config_overrides: Optional[dict] = None,
    ) -> ITlsContext:
        """Get an instance of the configured implementation of the tls context

        Args:
            session:  Session
                The current deploy session
            config_overrides:  Optional[dict]
                Optional runtime config values. These will overwrite any values
                pulled from the session.config

        Returns:
            tls_context:  ITlsContext
                The constructed instance of the context
        """
        # Check to see if this instance already exists
        if cls._instance is None or cls._instance_deploy_id != session.id:
            log.debug("Constructing TlsContext for [%s]", session.id)

            # Get the config
            tls_config = merge_configs(
                session.config.get(cls._APP_CONFIG_SECTION),
                config_overrides or {},
            )
            assert_config(
                tls_config is not None,
                f"Missing required config section: {cls._APP_CONFIG_SECTION}",
            )
            type_label = tls_config.get(cls._CONFIG_TYPE_FIELD)
            type_class = cls._registered_types.get(type_label)
            assert_config(
                type_class is not None,
                f"Cannot construct unknown TlsContext type [{type_label}]",
            )

            # Construct the instance
            cls._instance = type_class(session=session, config=tls_config)
            cls._instance_deploy_id = session.id

        log.debug2("Returning TlsContext for [%s]", session.id)
        return cls._instance

    @classmethod
    def register(cls, context_class: Type[ITlsContext]):
        """Register a constructor for a given context implementation type

        Args:
            context_class:  Type[ITlsContext]
                The ITlsContext child class to register
        """
        assert hasattr(context_class, ITlsContext._TYPE_LABEL_ATTRIBUTE), (
            "All derived ITlsContext classes must have an attribute "
            f"{ITlsContext._TYPE_LABEL_ATTRIBUTE}"
        )
        type_label = getattr(context_class, ITlsContext._TYPE_LABEL_ATTRIBUTE)
        if type_label in cls._registered_types:
            log.warning(
                "Received non-unique %s for %s: %s",
                ITlsContext._TYPE_LABEL_ATTRIBUTE,
                context_class,
                type_label,
            )
        log.debug2("Registering tls context type [%s]", type_label)
        cls._registered_types[type_label] = context_class



"""
Tests for the WatchThread
"""
# Standard
from unittest.mock import patch
import time

# Third Party
import pytest

# Local
from oper8.deploy_manager.dry_run_deploy_manager import DryRunDeployManager
from oper8.deploy_manager.kube_event import KubeEventType
from oper8.test_helpers.helpers import MockDeployManager, library_config
from oper8.test_helpers.pwm_helpers import (
    DisabledLeadershipManager,
    MockedReconcileThread,
    clear_caches,
    make_ownerref,
    make_resource,
)
from oper8.watch_manager.python_watch_manager.filters.filters import DisableFilter
from oper8.watch_manager.python_watch_manager.threads.watch import WatchThread
from oper8.watch_manager.python_watch_manager.utils.types import (
    ReconcileRequestType,
    ResourceId,
    WatchRequest,
)

## Helpers #####################################################################


@pytest.mark.timeout(5)
def test_watch_thread_happy_path():
    dm = DryRunDeployManager()
    watched_object = make_resource(spec={"test": "value"})
    watched_object_id = ResourceId.from_resource(watched_object)

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": None}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        watch_thread.start_thread()

        request = WatchRequest(
            # Set watched and requester to the same
            watched=watched_object_id,
            requester=watched_object_id,
        )
        watch_thread.request_watch(request)

        dm.deploy([watched_object])
        watched_object["spec"] = {"test": "updated"}
        dm.deploy([watched_object])

        time.sleep(1.5)
        watch_thread.stop_thread()
        assert mocked_reconcile_thread.get_request().type == KubeEventType.ADDED
        assert mocked_reconcile_thread.get_request().type == KubeEventType.MODIFIED


@pytest.mark.timeout(5)
def test_watch_thread_filtered():
    dm = DryRunDeployManager()
    watched_object = make_resource(spec={"test": "value"})
    watched_object_id = ResourceId.from_resource(watched_object)

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": DisableFilter}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        watch_thread.start_thread()

        request = WatchRequest(
            # Set watched and requester to the same
            watched=watched_object_id,
            requester=watched_object_id,
        )
        watch_thread.request_watch(request)

        dm.deploy([watched_object])
        watched_object["spec"] = {"test": "updated"}
        dm.deploy([watched_object])

        time.sleep(1.5)
        watch_thread.stop_thread()
        assert mocked_reconcile_thread.requests.empty()


@pytest.mark.timeout(5)
def test_watch_thread_deleted():
    dm = DryRunDeployManager()
    watched_object = make_resource(spec={"test": "value"})
    watched_object_id = ResourceId.from_resource(watched_object)

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": None}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        watch_thread.start_thread()

        request = WatchRequest(
            # Set watched and requester to the same
            watched=watched_object_id,
            requester=watched_object_id,
        )
        watch_thread.request_watch(request)

        dm.deploy([watched_object])
        dm.disable([watched_object])

        time.sleep(1.5)
        watch_thread.stop_thread()
        assert mocked_reconcile_thread.get_request().type == KubeEventType.ADDED
        assert mocked_reconcile_thread.get_request().type == KubeEventType.DELETED


@pytest.mark.timeout(5)
def test_watch_thread_owner_watch():
    dm = DryRunDeployManager()
    owner_object = make_resource(
        kind="DifferentKind", name="owner", spec={"test": "value"}
    )
    owner_object_id = ResourceId.from_resource(owner_object)
    watched_object = make_resource(
        spec={"test": "value"}, owner_refs=[make_ownerref(owner_object)]
    )
    watched_object_id = ResourceId.from_resource(watched_object)

    # Deploy owner before watch has started
    dm.deploy([owner_object])

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": None}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        request = WatchRequest(
            # Set watched and requester to the same
            watched=watched_object_id,
            requester=owner_object_id,
        )
        watch_thread.request_watch(request)
        watch_thread.start_thread()

        dm.deploy([watched_object])
        watched_object["spec"] = {"test": "updated"}
        dm.deploy([watched_object])

        time.sleep(1.5)
        watch_thread.stop_thread()
        assert (
            mocked_reconcile_thread.get_request().type == ReconcileRequestType.DEPENDENT
        )
        assert (
            mocked_reconcile_thread.get_request().type == ReconcileRequestType.DEPENDENT
        )


@pytest.mark.timeout(5)
def test_watch_thread_global_watch():
    dm = DryRunDeployManager()
    owner_object = make_resource(
        kind="DifferentKind", name="owner", spec={"test": "value"}
    )
    watched_object = make_resource(
        spec={"test": "value"}, owner_refs=[make_ownerref(owner_object)]
    )
    watched_object_id = ResourceId.from_resource(watched_object)

    dm.deploy([owner_object])

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": None}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        watch_thread.start_thread()

        request = WatchRequest(
            # Set watched and requester to the same
            watched=watched_object_id,
            requester=ResourceId(
                api_version=owner_object.get("apiVersion"),
                kind=owner_object.get("kind"),
            ),
        )
        watch_thread.request_watch(request)

        dm.deploy([watched_object])
        watched_object["spec"] = {"test": "updated"}
        dm.deploy([watched_object])

        time.sleep(3)
        watch_thread.stop_thread()
        assert (
            mocked_reconcile_thread.get_request().type == ReconcileRequestType.DEPENDENT
        )
        assert (
            mocked_reconcile_thread.get_request().type == ReconcileRequestType.DEPENDENT
        )


@pytest.mark.timeout(5)
def test_watch_thread_all_events():
    dm = DryRunDeployManager()
    watched_object = make_resource(spec={"test": "value"})
    request_resource_id = ResourceId(
        api_version=watched_object.get("apiVersion"), kind=watched_object.get("kind")
    )

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": None}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        watch_thread.start_thread()

        request = WatchRequest(
            # Set watched and requester to the same
            watched=request_resource_id,
            requester=request_resource_id,
        )
        watch_thread.request_watch(request)

        dm.deploy([watched_object])
        watched_object["spec"] = {"test": "updated"}
        dm.deploy([watched_object])
        dm.deploy([make_resource(name="second_obj")])
        dm.disable([watched_object])

        time.sleep(1.5)
        watch_thread.stop_thread()
        assert mocked_reconcile_thread.get_request().type == KubeEventType.ADDED
        assert mocked_reconcile_thread.get_request().type == KubeEventType.MODIFIED
        assert mocked_reconcile_thread.get_request().type == KubeEventType.ADDED
        assert mocked_reconcile_thread.get_request().type == KubeEventType.DELETED


@pytest.mark.timeout(5)
def test_watch_thread_global_watch_two_owners():
    dm = DryRunDeployManager()
    owner_object = make_resource(kind="OwnerKind", name="owner", spec={"test": "value"})
    owner_2_object = make_resource(
        kind="OwnerKind", name="owner2", spec={"test": "value"}
    )
    watched_object = make_resource(
        spec={"test": "value"},
        owner_refs=[make_ownerref(owner_object), make_ownerref(owner_2_object)],
    )
    watched_object_id = ResourceId.from_resource(watched_object)

    dm.deploy([owner_object])
    dm.deploy([owner_2_object])

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": None}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        request = WatchRequest(
            # Set watched and requester to the same
            watched=watched_object_id,
            requester=ResourceId(
                api_version=owner_object.get("apiVersion"),
                kind=owner_object.get("kind"),
            ),
        )
        watch_thread.request_watch(request)
        watch_thread.start_thread()

        dm.deploy([watched_object])
        watched_object["spec"] = {"test": "updated"}
        dm.deploy([watched_object])

        time.sleep(1.5)
        watch_thread.stop_thread()

        add_events = [
            mocked_reconcile_thread.get_request(),
            mocked_reconcile_thread.get_request(),
        ]
        assert "owner" in [event.resource.name for event in add_events]
        assert "owner2" in [event.resource.name for event in add_events]
        assert add_events[0].type == ReconcileRequestType.DEPENDENT
        assert add_events[1].type == ReconcileRequestType.DEPENDENT

        modified_events = [
            mocked_reconcile_thread.get_request(),
            mocked_reconcile_thread.get_request(),
        ]

        assert "owner" in [event.resource.name for event in modified_events]
        assert "owner2" in [event.resource.name for event in modified_events]
        assert modified_events[0].type == ReconcileRequestType.DEPENDENT
        assert modified_events[1].type == ReconcileRequestType.DEPENDENT


@pytest.mark.timeout(5)
def test_watch_thread_no_watch():
    dm = DryRunDeployManager()
    watched_object = make_resource(spec={"test": "value"})

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": DisableFilter}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        watch_thread.start_thread()

        dm.deploy([watched_object])
        watched_object["spec"] = {"test": "updated"}
        dm.deploy([watched_object])

        time.sleep(1.5)
        watch_thread.stop_thread()
        assert mocked_reconcile_thread.requests.empty()


@pytest.mark.timeout(5)
def test_watch_thread_not_leader():
    dm = DryRunDeployManager()
    watched_object = make_resource(spec={"test": "value"})
    watched_object_id = ResourceId.from_resource(watched_object)

    with library_config(
        python_watch_manager={"process_context": "fork", "filter": None}
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            leadership_manager=DisabledLeadershipManager(),
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )
        request = WatchRequest(watched=watched_object_id, requester=watched_object_id)
        watch_thread.request_watch(request)
        watch_thread.start_thread()

        dm.deploy([watched_object])
        watched_object["spec"] = {"test": "updated"}
        dm.deploy([watched_object])

        time.sleep(1.5)
        watch_thread.stop_thread()
        assert mocked_reconcile_thread.requests.empty()


@pytest.mark.timeout(5)
@pytest.mark.filterwarnings("ignore::pytest.PytestUnhandledThreadExceptionWarning")
def test_watch_thread_invalid_rbac():
    dm = MockDeployManager(watch_raise=True)
    watched_object = make_resource(spec={"test": "value"})
    watched_object_id = ResourceId.from_resource(watched_object)

    with patch(
        "oper8.watch_manager.python_watch_manager.threads.watch.os._exit",
        side_effect=Exception("EndTest"),
    ) as exit_mock, library_config(
        python_watch_manager={
            "process_context": "fork",
            "filter": None,
            "watch_retry_count": 3,
            "watch_retry_delay": "0.1s",
        }
    ):
        mocked_reconcile_thread = MockedReconcileThread()
        watch_thread = WatchThread(
            reconcile_thread=mocked_reconcile_thread,
            kind="Foo",
            api_version="foo.bar.com/v1",
            namespace="test",
            deploy_manager=dm,
        )

        request = WatchRequest(watched=watched_object_id, requester=watched_object_id)
        watch_thread.request_watch(request)
        watch_thread.start_thread()

        # Wait for the retries
        time.sleep(1)

        # Assert we tried to watch 4 times (3 retries plus the initial)
        assert dm.watch_objects.call_count == 4
        assert exit_mock.called



"""
The common connection type for a Redis instance
"""

# Standard
from typing import Optional, Tuple

# First Party
import alog

# Local
from .... import Session, assert_cluster, assert_precondition
from ....exceptions import ConfigError
from ...utils import common
from ..connection_base import DatastoreConnectionBase

## RedisConnection ###########################################################

log = alog.use_channel("RDS")


class RedisConnection(DatastoreConnectionBase):
    """
    A RedisConnection holds the core connection information for a named Redis
    instance, regardless of what IRedisComponent implements it. The key pieces
    of information are:

    * General config
        * hostname: The hostname where the instance can be reached
        * port: The port where the instance is listening

    * Auth
        * auth_secret_name: The in-cluster name for the secret holding the
            username and password
        * auth_secret_username_field: The field within the auth secret that
            holds the username.
        * auth_secret_password_field: The field within the auth secret that
            holds the password

    * TLS:
        * tls_cert: The content of the TLS cert if tls is enabled
        * tls_secret_name: The in-cluster name for the secret holding the TLS
            creds if tls is enabled
        * tls_secret_cert_field: The field within the tls secret that holds the
            cert
    """

    def __init__(
        self,
        session: Session,
        hostname: str,
        port: int,
        auth_secret_name: str,
        auth_secret_password_field: str,
        auth_secret_username_field: str,
        tls_secret_name: Optional[str] = None,
        tls_secret_cert_field: Optional[str] = None,
        auth_username: Optional[str] = None,
        auth_password: Optional[str] = None,
        tls_cert: Optional[str] = None,
    ):
        super().__init__(session)

        # These fields must be passed in directly
        self._hostname = hostname
        self._port = port
        self._auth_secret_name = auth_secret_name
        self._auth_secret_username_field = auth_secret_username_field
        self._auth_secret_password_field = auth_secret_password_field
        self._tls_secret_name = tls_secret_name
        self._tls_secret_cert_field = tls_secret_cert_field

        # The secret content may be populated or not, depending on whether this
        # Connection is being created by the component or a CR config. If not
        # populated now, they will be lazily populated on client request.
        self._auth_username = auth_username
        self._auth_password = auth_password
        self._tls_cert = tls_cert

        # Ensure that the TLS arguments are provided in a reasonable way. The
        # cert may be omitted
        tls_args = {tls_secret_name, tls_secret_cert_field}
        assert (
            tls_args == {None} or None not in tls_args
        ), "All TLS arguments must be provided if tls is enabled"
        self._tls_enabled = tls_args != {None}
        assert (
            self._tls_enabled or self._tls_cert is None
        ), "Cannot give a tls cert value when tls is disabled"

        # Schema is deduced based on the presence of the tls arguments
        self._schema = "redis" if tls_secret_name is None else "rediss"

    ## Properties ##############################################################

    @property
    def schema(self) -> str:
        """The schema (redis or rediss)"""
        return self._schema

    @property
    def hostname(self) -> str:
        """The hostname (without schema)"""
        return self._hostname

    @property
    def port(self) -> int:
        """The numeric port"""
        return self._port

    @property
    def auth_secret_name(self) -> str:
        """Secret name containing the username_key and password_key"""
        return self._auth_secret_name

    @property
    def auth_secret_username_field(self) -> str:
        """Field in the auth secret containing the username"""
        return self._auth_secret_username_field

    @property
    def auth_secret_password_field(self) -> str:
        """Field in the auth secret containing the password"""
        return self._auth_secret_password_field

    @property
    def tls_secret_name(self) -> str:
        """The name of the secret holding the tls certificate (for mounting)"""
        return self._tls_secret_name

    @property
    def tls_secret_cert_field(self) -> str:
        """The field within the tls secret that holds the CA cert"""
        return self._tls_secret_cert_field

    @property
    def tls_enabled(self) -> bool:
        return self._tls_enabled

    ## Interface ###############################################################

    _DICT_FIELDS = [
        "hostname",
        "port",
        "auth_secret_name",
        "auth_secret_password_field",
        "auth_secret_username_field",
        "tls_secret_name",
        "tls_secret_cert_field",
    ]

    def to_dict(self) -> dict:
        """Return the dict representation of the object for the CR"""
        return {field: getattr(self, f"_{field}") for field in self._DICT_FIELDS}

    @classmethod
    def from_dict(cls, session: Session, config_dict: dict) -> "RedisConnection":
        kwargs = {"session": session}
        config_dict = common.camelcase_to_snake_case(config_dict)
        uri_secret = config_dict.get("uri_secret")
        uri_hostname_field = config_dict.get("uri_secret_hostname_field")
        uri_port_field = config_dict.get("uri_secret_port_field")

        # First pull provided hostname/port secret if available and fill in
        # hostname/port fields into config_dict
        if uri_secret and uri_hostname_field and uri_port_field:
            # If we have provided host/port credentials, we need to extract them
            # and place these values in our config dict
            success, secret_content = session.get_object_current_state(
                "Secret", uri_secret
            )
            assert_cluster(success, f"Fetching connection secret [{uri_secret}] failed")
            assert "data" in secret_content, "Got a secret without 'data'?"
            secret_content = secret_content.get("data")
            assert_precondition(
                secret_content,
                f"Missing expected Secret/{uri_secret} holding [hostname] and [port]",
            )
            hostname = common.b64_secret_decode(secret_content.get(uri_hostname_field))
            port = common.b64_secret_decode(secret_content.get(uri_port_field))
            if None in [hostname, port]:
                log.debug2(
                    "Failed to find hostname/port in uri secret [%s]",
                    uri_secret,
                )

            try:
                port = int(port)
            except ValueError as err:
                raise ConfigError(f"Invalid non-int port: {port}") from err
            config_dict["hostname"], config_dict["port"] = hostname, port

        for field in cls._DICT_FIELDS:
            if field not in config_dict:
                raise ValueError(f"Missing required connection element [{field}]")

            # Set the kwargs (using None in place of empty strings)
            kwargs[field] = config_dict[field] or None

        return cls(**kwargs)

    ## Client Utilities ########################################################

    def get_auth_username_password(self) -> Tuple[Optional[str], Optional[str]]:
        """Get the current username_key/password_key pair from the auth secret if
        available

        Returns:
            username:  str or None
                The plain-text username (not encoded) if available
            password:  str or None
                The plain-text password (not encoded) if available
        """
        if None in [self._auth_username, self._auth_password]:
            secret_content = self._fetch_secret_data(self._auth_secret_name) or {}
            log.debug4("Auth secret content: %s", secret_content)
            log.debug3(
                "Looking for [%s/%s]",
                self._auth_secret_username_field,
                self._auth_secret_password_field,
            )
            username = secret_content.get(self._auth_secret_username_field)
            password = secret_content.get(self._auth_secret_password_field)
            # username not required as expect username only used when using ACL
            # redis-cli also does support username in URI to be used for that
            # CITE: https://redis.io/commands/auth
            if None in [username, password]:
                log.debug2(
                    "Failed to find username/password in auth secret [%s]",
                    self._auth_secret_name,
                )
                return None, None
            self._auth_username = username
            self._auth_password = password
        return self._auth_username, self._auth_password

    def get_tls_cert(self) -> Optional[str]:
        """Get the current TLS certificate for a client connection if TLS is
        enabled

        If TLS is enabled, but the cert is not found, this function will raise
        an AssertionError

        Returns:
            tls_cert: str or None
                PEM encoded cert string (not base64-encoded) if found, otherwise
                None
        """
        if self._tls_enabled:
            if self._tls_cert is None:
                secret_data = self._fetch_secret_data(self._tls_secret_name)
                if secret_data is not None:
                    self._tls_cert = secret_data.get(self._tls_secret_cert_field)
            return self._tls_cert

        return None

    def get_connection_string(self) -> str:
        """Get the formatted Redis connection string to connect to the instance

        Returns:
            connection_string:  str
                The formatted connection string
        """
        username_key, password_key = self.get_auth_username_password()
        assert_precondition(
            None not in [username_key, password_key],
            "No auth keys available for Redis connection string",
        )
        # NOTE: username/password required and needs to change if ever need to
        # support the rediss://<host>:<port> format without username/password
        # CITE: https://redis.io/topics/rediscli
        return "{}://{}:{}@{}:{}".format(
            self._schema, username_key, password_key, self._hostname, self._port
        )



"""Implementation of the Leader-for-Life LeaderElection"""
# First Party
import alog

# Local
from .... import config
from ....deploy_manager.owner_references import update_owner_references
from ....exceptions import ConfigError, assert_config
from ....managed_object import ManagedObject
from ....utils import nested_get
from ..utils import get_operator_namespace, get_pod_name
from .base import ThreadedLeaderManagerBase

log = alog.use_channel("LDRLIFE")


class LeaderForLifeManager(ThreadedLeaderManagerBase):
    """
    LeaderForLifeManager Class implements the old "leader-for-life" operator-sdk
    lock type. This lock creates a configmap with the operator pod as owner in
    the current namespace. This way when the pod is deleted or list so is the
    configmap.
    """

    def __init__(self, deploy_manager):
        """
        Initialize class with lock_name, current namespace, and pod information
        """
        super().__init__(deploy_manager)

        # Gather lock_name, namespace and pod manifest
        self.lock_name = (
            config.operator_name
            if config.operator_name
            else config.python_watch_manager.lock.name
        )

        self.namespace = get_operator_namespace()
        pod_name = get_pod_name()
        assert_config(self.lock_name, "Unable to detect lock name")
        assert_config(self.namespace, "Unable to detect operator namespace")
        assert_config(pod_name, "Unable to detect pod name")

        # Get the current pod context which is used in the lock configmap
        log.debug("Gathering pod context information")
        success, pod_obj = self.deploy_manager.get_object_current_state(
            kind="Pod", name=pod_name, namespace=self.namespace, api_version="v1"
        )
        if not success or not pod_obj:
            log.error(
                "Unable to fetch pod %s/%s Unable to use leader-for-life without ownerReference",
                self.namespace,
                pod_name,
            )
            raise ConfigError(
                f"Unable to fetch pod {self.namespace}/{pod_name}."
                "Unable to use leader-for-life without ownerReference"
            )

        self.pod_manifest = ManagedObject(pod_obj)

    def renew_or_acquire(self):
        """
        Renew or acquire lock by checking the current configmap status
        """
        # Get current config map
        success, cluster_config_map = self.deploy_manager.get_object_current_state(
            kind="ConfigMap",
            name=self.lock_name,
            namespace=self.namespace,
            api_version="v1",
        )
        if not success:
            log.warning(
                "Unable to fetch config map %s/%s", self.namespace, self.lock_name
            )

        # If configmap exists then verify owner ref
        if cluster_config_map:
            log.debug2(
                f"ConfigMap Lock {cluster_config_map} already exists, checking ownership"
            )
            owner_ref_list = nested_get(
                cluster_config_map, "metadata.ownerReferences", []
            )
            if len(owner_ref_list) != 1:
                log.error(
                    "Invalid leadership config map detected. Only one owner allowed"
                )
                self.release_lock()
                return

            if owner_ref_list[0].get("uid") == self.pod_manifest.uid:
                self.acquire_lock()
            else:
                self.release_lock()

        # Create configmap if it doesn't exist
        else:
            log.debug2(f"ConfigMap Lock {cluster_config_map} does not exist, creating")
            config_map = {
                "kind": "ConfigMap",
                "apiVersion": "v1",
                "metadata": {
                    "name": self.lock_name,
                    "namespace": self.namespace,
                },
            }
            update_owner_references(
                self.deploy_manager, self.pod_manifest.definition, config_map
            )
            success, _ = self.deploy_manager.deploy(
                [config_map], manage_owner_references=False
            )
            if not success:
                log.warning("Unable to acquire leadership lock")
                self.release_lock()
            else:
                self.acquire_lock()
